# -*- coding: utf-8 -*-
"""
Created on Mon Mar 14 15:33:07 2016

@author: timothyb0912
         José Ángel Martín Baos
@module: pylogit
"""
from __future__ import absolute_import

from .pylogit import create_choice_model
from .bootstrap import Boot
from .choice_tools import convert_wide_to_long
from .choice_tools import convert_long_to_wide

from .choice_tools import divide_long_train_test
from .choice_tools import KFold_cross_val
from .kernel_logistic_regression import *



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 22 08:32:48 2016

@module:    choice_tools.py
@name:      Helpful Tools for Choice Model Estimation
@author:    Timothy Brathwaite
            José Ángel Martín Baos
@summary:   Contains functions that help prepare one's data for choice model
            estimation or helps speed the estimation process (the 'mappings').
"""
import warnings
from collections import OrderedDict
from collections import Iterable
from numbers import Number

import numpy as np
import pandas as pd
import random

from scipy.sparse import csr_matrix
from scipy.sparse import issparse


def get_dataframe_from_data(data):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV file
        containing the long format data for this choice model. Note long format
        has one row per available alternative for each observation. If pandas
        dataframe, the dataframe should be the long format data for the choice
        model.

    Returns
    -------
    dataframe : pandas dataframe of the long format data for the choice model.
    """
    if isinstance(data, str):
        if data.endswith(".csv"):
            dataframe = pd.read_csv(data)
        else:
            msg_1 = "data = {} is of unknown file type."
            msg_2 = " Please pass path to csv."
            raise ValueError(msg_1.format(data) + msg_2)
    elif isinstance(data, pd.DataFrame):
        dataframe = data
    else:
        msg_1 = "type(data) = {} is an invalid type."
        msg_2 = " Please pass pandas dataframe or path to csv."
        raise TypeError(msg_1.format(type(data)) + msg_2)

    return dataframe


# In a multinomial choice model, all data should be converted to "long format"
# Long format has one row per observation per available alternative for each
# person.

# Each row should have the same number of columns, so all utility coefficients
# must be part of each utility specification.

# This means that the variables which belong to coefficients that are not
# "naturally" in all utility equations must be redefined so that the
# variable == 0 when the utility equation does not correspond to the beta that
# is associated with the variable.

# Additionally, variables which show up in all utility equations but have
# different betas in all utility equations should be turned into separate
# variables--one per utility equation-- where the variable == 0 if the
# alternative does not correspond to the beta for that variable.


def ensure_object_is_ordered_dict(item, title):
    """
    Checks that the item is an OrderedDict. If not, raises ValueError.
    """
    assert isinstance(title, str)

    if not isinstance(item, OrderedDict):
        msg = "{} must be an OrderedDict. {} passed instead."
        raise TypeError(msg.format(title, type(item)))

    return None


def ensure_object_is_string(item, title):
    """
    Checks that the item is a string. If not, raises ValueError.
    """
    assert isinstance(title, str)

    if not isinstance(item, str):
        msg = "{} must be a string. {} passed instead."
        raise TypeError(msg.format(title, type(item)))

    return None


def ensure_object_is_ndarray(item, title):
    """
    Ensures that a given mapping matrix is a dense numpy array. Raises a
    helpful TypeError if otherwise.
    """
    assert isinstance(title, str)

    if not isinstance(item, np.ndarray):
        msg = "{} must be a np.ndarray. {} passed instead."
        raise TypeError(msg.format(title, type(item)))

    return None


def ensure_columns_are_in_dataframe(columns,
                                    dataframe,
                                    col_title='',
                                    data_title='data'):
    """
    Checks whether each column in `columns` is in `dataframe`. Raises
    ValueError if any of the columns are not in the dataframe.

    Parameters
    ----------
    columns : list of strings.
        Each string should represent a column heading in dataframe.
    dataframe : pandas DataFrame.
        Dataframe containing the data for the choice model to be estimated.
    col_title : str, optional.
        Denotes the title of the columns that were passed to the function.
    data_title : str, optional.
        Denotes the title of the dataframe that is being checked to see whether
        it contains the passed columns. Default == 'data'

    Returns
    -------
    None.
    """
    # Make sure columns is an iterable
    assert isinstance(columns, Iterable)
    # Make sure dataframe is a pandas dataframe
    assert isinstance(dataframe, pd.DataFrame)
    # Make sure title is a string
    assert isinstance(col_title, str)
    assert isinstance(data_title, str)

    problem_cols = [col for col in columns if col not in dataframe.columns]
    if problem_cols != []:
        if col_title == '':
            msg = "{} not in {}.columns"
            final_msg = msg.format(problem_cols, data_title)
        else:
            msg = "The following columns in {} are not in {}.columns: {}"
            final_msg = msg.format(col_title, data_title, problem_cols)

        raise ValueError(final_msg)

    return None


def check_argument_type(long_form, specification_dict):
    """
    Ensures that long_form is a pandas dataframe and that specification_dict
    is an OrderedDict, raising a ValueError otherwise.

    Parameters
    ----------
    long_form : pandas dataframe.
        Contains one row for each available alternative, for each observation.
    specification_dict : OrderedDict.
        Keys are a proper subset of the columns in `long_form_df`. Values are
        either a list or a single string, `"all_diff"` or `"all_same"`. If a
        list, the elements should be:

            - single objects that are within the alternative ID column of
              `long_form_df`
            - lists of objects that are within the alternative ID column of
              `long_form_df`. For each single object in the list, a unique
              column will be created (i.e. there will be a unique coefficient
              for that variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification_dict` values, a single column will be created for
              all the alternatives within iterable (i.e. there will be one
              common coefficient for the variables in the iterable).

    Returns
    -------
    None.
    """
    if not isinstance(long_form, pd.DataFrame):
        msg = "long_form should be a pandas dataframe. It is a {}"
        raise TypeError(msg.format(type(long_form)))

    ensure_object_is_ordered_dict(specification_dict, "specification_dict")

    return None


def ensure_alt_id_in_long_form(alt_id_col, long_form):
    """
    Ensures alt_id_col is in long_form, and raises a ValueError if not.

    Parameters
    ----------
    alt_id_col : str.
        Column name which denotes the column in `long_form` that contains the
        alternative ID for each row in `long_form`.
    long_form : pandas dataframe.
        Contains one row for each available alternative, for each observation.

    Returns
    -------
    None.
    """
    if alt_id_col not in long_form.columns:
        msg = "alt_id_col == {} is not a column in long_form."
        raise ValueError(msg.format(alt_id_col))

    return None


def ensure_specification_cols_are_in_dataframe(specification, dataframe):
    """
    Checks whether each column in `specification` is in `dataframe`. Raises
    ValueError if any of the columns are not in the dataframe.

    Parameters
    ----------
    specification : OrderedDict.
        Keys are a proper subset of the columns in `data`. Values are either a
        list or a single string, "all_diff" or "all_same". If a list, the
        elements should be:
            - single objects that are in the alternative ID column of `data`
            - lists of objects that are within the alternative ID column of
              `data`. For each single object in the list, a unique column will
              be created (i.e. there will be a unique coefficient for that
              variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification` values, a single column will be created for all
              the alternatives within the iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    dataframe : pandas DataFrame.
        Dataframe containing the data for the choice model to be estimated.

    Returns
    -------
    None.
    """
    # Make sure specification is an OrderedDict
    try:
        assert isinstance(specification, OrderedDict)
    except AssertionError:
        raise TypeError("`specification` must be an OrderedDict.")
    # Make sure dataframe is a pandas dataframe
    assert isinstance(dataframe, pd.DataFrame)

    problem_cols = []
    dataframe_cols = dataframe.columns
    for key in specification:
        if key not in dataframe_cols:
            problem_cols.append(key)
    if problem_cols != []:
        msg = "The following keys in the specification are not in 'data':\n{}"
        raise ValueError(msg.format(problem_cols))

    return None


def check_type_and_values_of_specification_dict(specification_dict,
                                                unique_alternatives):
    """
    Verifies that the values of specification_dict have the correct type, have
    the correct structure, and have valid values (i.e. are actually in the set
    of possible alternatives). Will raise various errors if / when appropriate.

    Parameters
    ----------
    specification_dict : OrderedDict.
        Keys are a proper subset of the columns in `long_form_df`. Values are
        either a list or a single string, `"all_diff"` or `"all_same"`. If a
        list, the elements should be:

            - single objects that are within the alternative ID column of
              `long_form_df`
            - lists of objects that are within the alternative ID column of
              `long_form_df`. For each single object in the list, a unique
              column will be created (i.e. there will be a unique coefficient
              for that variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification_dict` values, a single column will be created for
              all the alternatives within iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    unique_alternatives : 1D ndarray.
        Should contain the possible alternative id's for this dataset.

    Returns
    -------
    None.
    """
    for key in specification_dict:
        specification = specification_dict[key]
        if isinstance(specification, str):
            if specification not in ["all_same", "all_diff"]:
                msg = "specification_dict[{}] not in ['all_same', 'all_diff']"
                raise ValueError(msg.format(key))

        elif isinstance(specification, list):
            # Imagine that the specification is [[1, 2], 3]
            # group would be [1, 2]
            # group_item would be 1 or 2. group_item should never be a list.
            for group in specification:
                group_is_list = isinstance(group, list)
                if group_is_list:
                    for group_item in group:
                        if isinstance(group_item, list):
                            msg = "Wrong structure for specification_dict[{}]"
                            msg_2 = " Values can be a list of lists of ints,"
                            msg_3 = " not lists of lists of lists of ints."
                            total_msg = msg.format(key) + msg_2 + msg_3
                            raise ValueError(total_msg)

                        elif group_item not in unique_alternatives:
                            msg_1 = "{} in {} in specification_dict[{}]"
                            msg_2 = " is not in long_format[alt_id_col]"
                            total_msg = (msg_1.format(group_item, group, key) +
                                         msg_2)
                            raise ValueError(total_msg)
                else:
                    if group not in unique_alternatives:
                        msg_1 = "{} in specification_dict[{}]"
                        msg_2 = " is not in long_format[alt_id_col]"
                        raise ValueError(msg_1.format(group, key) + msg_2)

        else:
            msg = "specification_dict[{}] must be 'all_same', 'all_diff', or"
            msg_2 = " a list."
            raise TypeError(msg.format(key) + msg_2)

    return None


def check_keys_and_values_of_name_dictionary(names,
                                             specification_dict,
                                             num_alts):
    """
    Check the validity of the keys and values in the names dictionary.

    Parameters
    ----------
    names : OrderedDict, optional.
        Should have the same keys as `specification_dict`. For each key:

            - if the corresponding value in `specification_dict` is "all_same",
              then there should be a single string as the value in names.
            - if the corresponding value in `specification_dict` is "all_diff",
              then there should be a list of strings as the value in names.
              There should be one string in the value in names for each
              possible alternative.
            - if the corresponding value in `specification_dict` is a list,
              then there should be a list of strings as the value in names.
              There should be one string the value in names per item in the
              value in `specification_dict`.
    specification_dict : OrderedDict.
        Keys are a proper subset of the columns in `long_form_df`. Values are
        either a list or a single string, `"all_diff"` or `"all_same"`. If a
        list, the elements should be:

            - single objects that are within the alternative ID column of
              `long_form_df`
            - lists of objects that are within the alternative ID column of
              `long_form_df`. For each single object in the list, a unique
              column will be created (i.e. there will be a unique coefficient
              for that variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification_dict` values, a single column will be created for
              all the alternatives within iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    num_alts : int.
        The number of alternatives in this dataset's universal choice set.

    Returns
    -------
    None.
    """
    if names.keys() != specification_dict.keys():
        msg = "names.keys() does not equal specification_dict.keys()"
        raise ValueError(msg)

    for key in names:
        specification = specification_dict[key]
        name_object = names[key]
        if isinstance(specification, list):
            try:
                assert isinstance(name_object, list)
                assert len(name_object) == len(specification)
                assert all([isinstance(x, str) for x in name_object])
            except AssertionError:
                msg = "names[{}] must be a list AND it must have the same"
                msg_2 = " number of strings as there are elements of the"
                msg_3 = " corresponding list in specification_dict"
                raise ValueError(msg.format(key) + msg_2 + msg_3)

        else:
            if specification == "all_same":
                if not isinstance(name_object, str):
                    msg = "names[{}] should be a string".format(key)
                    raise TypeError(msg)

            else:    # This means speciffication == 'all_diff'
                try:
                    assert isinstance(name_object, list)
                    assert len(name_object) == num_alts
                except AssertionError:
                    msg_1 = "names[{}] should be a list with {} elements,"
                    msg_2 = " 1 element for each possible alternative"
                    msg = (msg_1.format(key, num_alts) + msg_2)
                    raise ValueError(msg)

    return None


def ensure_all_columns_are_used(num_vars_accounted_for,
                                dataframe,
                                data_title='long_data'):
    """
    Ensure that all of the columns from dataframe are in the list of used_cols.
    Will raise a helpful UserWarning if otherwise.

    Parameters
    ----------
    num_vars_accounted_for : int.
        Denotes the number of variables used in one's function.
    dataframe : pandas dataframe.
        Contains all of the data to be converted from one format to another.
    data_title : str, optional.
        Denotes the title by which `dataframe` should be referred in the
        UserWarning.

    Returns
    -------
    None.
    """
    dataframe_vars = set(dataframe.columns.tolist())
    num_dataframe_vars = len(dataframe_vars)

    if num_vars_accounted_for == num_dataframe_vars:
        pass

    elif num_vars_accounted_for < num_dataframe_vars:
        msg = "Note, there are {:,} variables in {} but the inputs"
        msg_2 = " ind_vars, alt_specific_vars, and subset_specific_vars only"
        msg_3 = " account for {:,} variables."

        warnings.warn(msg.format(num_dataframe_vars, data_title) +
                      msg_2 + msg_3.format(num_vars_accounted_for))

    else:  # This means num_vars_accounted_for > num_dataframe_vars
        msg = "There are more variable specified in ind_vars, "
        msg_2 = "alt_specific_vars, and subset_specific_vars ({:,}) than there"
        msg_3 = " are variables in {} ({:,})"
        warnings.warn(msg +
                      msg_2.format(num_vars_accounted_for) +
                      msg_3.format(data_title, num_dataframe_vars))

    return None


def check_dataframe_for_duplicate_records(obs_id_col, alt_id_col, df):
    """
    Checks a cross-sectional dataframe of long-format data for duplicate
    observations. Duplicate observations are defined as rows with the same
    observation id value and the same alternative id value.

    Parameters
    ----------
    obs_id_col : str.
        Denotes the column in `df` that contains the observation ID
        values for each row.
    alt_id_col : str.
        Denotes the column in `df` that contains the alternative ID
        values for each row.
    df : pandas dataframe.
        The dataframe of long format data that is to be checked for duplicates.

    Returns
    -------
    None.
    """
    if df.duplicated(subset=[obs_id_col, alt_id_col]).any():
        msg = "One or more observation-alternative_id pairs is not unique."
        raise ValueError(msg)

    return None


def ensure_num_chosen_alts_equals_num_obs(obs_id_col, choice_col, df):
    """
    Checks that the total number of recorded choices equals the total number of
    observations. If this is not the case, raise helpful ValueError messages.

    Parameters
    ----------
    obs_id_col : str.
        Denotes the column in `df` that contains the observation ID values for
        each row.
    choice_col : str.
        Denotes the column in `long_data` that contains a one if the
        alternative pertaining to the given row was the observed outcome for
        the observation pertaining to the given row and a zero otherwise.
    df : pandas dataframe.
        The dataframe whose choices and observations will be checked.

    Returns
    -------
    None.
    """
    num_obs = df[obs_id_col].unique().shape[0]
    num_choices = df[choice_col].sum()

    if num_choices < num_obs:
        msg = "One or more observations have not chosen one "
        msg_2 = "of the alternatives available to him/her"
        raise ValueError(msg + msg_2)
    if num_choices > num_obs:
        msg = "One or more observations has chosen multiple alternatives"
        raise ValueError(msg)

    return None


def check_type_and_values_of_alt_name_dict(alt_name_dict, alt_id_col, df):
    """
    Ensures that `alt_name_dict` is a dictionary and that its keys are in the
    alternative id column of `df`. Raises helpful errors if either condition
    is not met.

    Parameters
    ----------
    alt_name_dict : dict.
        A dictionary whose keys are the possible values in
        `df[alt_id_col].unique()`. The values should be the name that one
        wants to associate with each alternative id.
    alt_id_col : str.
        Denotes the column in `df` that contains the alternative ID values for
        each row.
    df : pandas dataframe.
        The dataframe of long format data that contains the alternative IDs.

    Returns
    -------
    None.
    """
    if not isinstance(alt_name_dict, dict):
        msg = "alt_name_dict should be a dictionary. Passed value was a {}"
        raise TypeError(msg.format(type(alt_name_dict)))

    if not all([x in df[alt_id_col].values for x in alt_name_dict.keys()]):
        msg = "One or more of alt_name_dict's keys are not "
        msg_2 = "in long_data[alt_id_col]"
        raise ValueError(msg + msg_2)

    return None


def ensure_PMLE_lambda_is_scalar_or_none(PMLE_lambda):
    """
    Ensures that `PMLE_lambda` is a scalar value. Raises a helpful
    TypeError otherwise.

    Parameters
    ----------
    PMLE_lambda : int, float or long
        Scalar value determining the PMLE penalty for the regression.

    Returns
    -------
    None.
    """
    if not isinstance(PMLE_lambda, Number):
        msg_1 = "PMLE_lambda should be an int, float, or long."
        msg_2 = "The passed value of PMLE_lambda had type: {}".format(type(PMLE_lambda))
        raise TypeError(msg_1 + msg_2)

    return None


def create_design_matrix(long_form,
                         specification_dict,
                         alt_id_col,
                         names=None):
    """
    Parameters
    ----------
    long_form : pandas dataframe.
        Contains one row for each available alternative, for each observation.
    specification_dict : OrderedDict.
        Keys are a proper subset of the columns in `long_form_df`. Values are
        either a list or a single string, `"all_diff"` or `"all_same"`. If a
        list, the elements should be:

            - single objects that are within the alternative ID column of
              `long_form_df`
            - lists of objects that are within the alternative ID column of
              `long_form_df`. For each single object in the list, a unique
              column will be created (i.e. there will be a unique coefficient
              for that variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification_dict` values, a single column will be created for
              all the alternatives within iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    alt_id_col : str.
        Column name which denotes the column in `long_form` that contains the
        alternative ID for each row in `long_form`.
    names : OrderedDict, optional.
        Should have the same keys as `specification_dict`. For each key:

            - if the corresponding value in `specification_dict` is "all_same",
              then there should be a single string as the value in names.
            - if the corresponding value in `specification_dict` is "all_diff",
              then there should be a list of strings as the value in names.
              There should be one string in the value in names for each
              possible alternative.
            - if the corresponding value in `specification_dict` is a list,
              then there should be a list of strings as the value in names.
              There should be one string the value in names per item in the
              value in `specification_dict`.
        Default == None.

    Returns
    -------
    design_matrix, var_names: tuple with two elements.
        First element is the design matrix, a numpy array with some number of
        columns and as many rows as are in `long_form`. Each column corresponds
        to a coefficient to be estimated. The second element is a list of
        strings denoting the names of each coefficient, with one variable name
        per column in the design matrix.
    """
    ##########
    # Check that the arguments meet this functions assumptions.
    # Fail gracefully if the arguments do not meet the function's requirements.
    #########
    check_argument_type(long_form, specification_dict)

    ensure_alt_id_in_long_form(alt_id_col, long_form)

    ensure_specification_cols_are_in_dataframe(specification_dict, long_form)

    # Find out what and how many possible alternatives there are
    unique_alternatives = np.sort(long_form[alt_id_col].unique())
    num_alternatives = len(unique_alternatives)

    check_type_and_values_of_specification_dict(specification_dict,
                                                unique_alternatives)

    # Check the user passed dictionary of names if the user passed such a list
    if names is not None:
        ensure_object_is_ordered_dict(names, "names")

        check_keys_and_values_of_name_dictionary(names,
                                                 specification_dict,
                                                 num_alternatives)

    ##########
    # Actually create the design matrix
    ##########
    # Create a list of the columns of independent variables
    independent_vars = []
    # Create a list of variable names
    var_names = []

    # Create the columns of the design matrix based on the specification dict.
    for variable in specification_dict:
        specification = specification_dict[variable]
        if specification == "all_same":
            # Create the variable column
            independent_vars.append(long_form[variable].values)
            # Create the column name
            var_names.append(variable)
        elif specification == "all_diff":
            for alt in unique_alternatives:
                # Create the variable column
                independent_vars.append((long_form[alt_id_col] == alt).values *
                                        long_form[variable].values)
                # create the column name
                var_names.append("{}_{}".format(variable, alt))
        else:
            for group in specification:
                if isinstance(group, list):
                    # Create the variable column
                    independent_vars.append(
                                     long_form[alt_id_col].isin(group).values *
                                     long_form[variable].values)
                    # Create the column name
                    var_names.append("{}_{}".format(variable, str(group)))

                else:  # the group is an integer
                    # Create the variable column
                    new_col_vals = ((long_form[alt_id_col] == group).values *
                                    long_form[variable].values)
                    independent_vars.append(new_col_vals)
                    # Create the column name
                    var_names.append("{}_{}".format(variable, group))


    design_matrix_as_list = []
    for x in independent_vars:
        design_matrix_as_list.append(x[:, None])

    # Create the final design matrix
    design_matrix = np.hstack(design_matrix_as_list)

    # Use the list of names passed by the user, if the user passed such a list
    if names is not None:
        var_names = []
        for value in names.values():
            if isinstance(value, str):
                var_names.append(value)
            else:
                for inner_name in value:
                    var_names.append(inner_name)

    return design_matrix, var_names


def get_original_order_unique_ids(id_array):
    """
    Get the unique id's of id_array, in their original order of appearance.

    Parameters
    ----------
    id_array : 1D ndarray.
        Should contain the ids that we want to extract the unique values from.

    Returns
    -------
    original_order_unique_ids : 1D ndarray.
        Contains the unique ids from `id_array`, in their original order of
        appearance.
    """
    assert isinstance(id_array, np.ndarray)
    assert len(id_array.shape) == 1

    # Get the indices of the unique IDs in their order of appearance
    # Note the [1] is because the np.unique() call will return both the sorted
    # unique IDs and the indices
    original_unique_id_indices =\
        np.sort(np.unique(id_array, return_index=True)[1])

    # Get the unique ids, in their original order of appearance
    original_order_unique_ids = id_array[original_unique_id_indices]

    return original_order_unique_ids


def create_row_to_some_id_col_mapping(id_array):
    """
    Parameters
    ----------
    id_array : 1D ndarray.
        All elements of the array should be ints representing some id related
        to the corresponding row.

    Returns
    -------
    rows_to_ids : 2D scipy sparse array.
        Will map each row of id_array to the unique values of `id_array`. The
        columns of the returned sparse array will correspond to the unique
        values of `id_array`, in the order of appearance for each of these
        unique values.
    """
    # Get the unique ids, in their original order of appearance
    original_order_unique_ids = get_original_order_unique_ids(id_array)

    # Create a matrix with the same number of rows as id_array but a single
    # column for each of the unique IDs. This matrix will associate each row
    # as belonging to a particular observation using a one and using a zero to
    # show non-association.
    rows_to_ids = (id_array[:, None] ==
                   original_order_unique_ids[None, :]).astype(int)
    return rows_to_ids


def create_sparse_mapping(id_array, unique_ids=None):
    """
    Will create a scipy.sparse compressed-sparse-row matrix that maps
    each row represented by an element in id_array to the corresponding
    value of the unique ids in id_array.

    Parameters
    ----------
    id_array : 1D ndarray of ints.
        Each element should represent some id related to the corresponding row.
    unique_ids : 1D ndarray of ints, or None, optional.
        If not None, each element should be present in `id_array`. The elements
        in `unique_ids` should be present in the order in which one wishes them
        to appear in the columns of the resulting sparse array. For the
        `row_to_obs` and `row_to_mixers` mappings, this should be the order of
        appearance in `id_array`. If None, then the unique_ids will be created
        from `id_array`, in the order of their appearance in `id_array`.

    Returns
    -------
    mapping : 2D scipy.sparse CSR matrix.
        Will contain only zeros and ones. `mapping[i, j] == 1` where
        `id_array[i] == unique_ids[j]`. The id's corresponding to each column
        are given by `unique_ids`. The rows correspond to the elements of
        `id_array`.
    """
    # Create unique_ids if necessary
    if unique_ids is None:
        unique_ids = get_original_order_unique_ids(id_array)

    # Check function arguments for validity
    assert isinstance(unique_ids, np.ndarray)
    assert isinstance(id_array, np.ndarray)
    assert unique_ids.ndim == 1
    assert id_array.ndim == 1

    # Figure out which ids in id_array are represented in unique_ids
    represented_ids = np.in1d(id_array, unique_ids)
    # Determine the number of rows in id_array that are in unique_ids
    num_non_zero_rows = represented_ids.sum()
    # Figure out the dimensions of the resulting sparse matrix
    num_rows = id_array.size
    num_cols = unique_ids.size
    # Specify the non-zero values that will be present in the sparse matrix.
    data = np.ones(num_non_zero_rows, dtype=int)
    # Specify which rows will have non-zero entries in the sparse matrix.
    row_indices = np.arange(num_rows)[represented_ids]
    # Map the unique id's to their respective columns
    unique_id_dict = dict(zip(unique_ids, np.arange(num_cols)))
    # Figure out the column indices of the non-zero entries, and do so in a way
    # that avoids a key error (i.e. only look up ids that are represented)
    col_indices =\
        np.array([unique_id_dict[x] for x in id_array[represented_ids]])

    # Create and return the sparse matrix
    return csr_matrix((data, (row_indices, col_indices)),
                      shape=(num_rows, num_cols))


##########
# Create a function to create the mapping matrices from long form rows to the
# associated observations, from long form rows to alternative IDs, and
# from "chosen" long form rows to the associated observations
##########
def create_long_form_mappings(long_form,
                              obs_id_col,
                              alt_id_col,
                              choice_col=None,
                              nest_spec=None,
                              mix_id_col=None,
                              dense=False):
    """
    Parameters
    ----------
    long_form : pandas dataframe.
        Contains one row for each available alternative for each observation.
    obs_id_col : str.
        Denotes the column in `long_form` which contains the choice situation
        observation ID values for each row of `long_form`. Note each value in
        this column must be unique (i.e., individuals with repeat observations
        have unique `obs_id_col` values for each choice situation, and
        `obs_id_col` values are unique across individuals).
    alt_id_col : str.
        Denotes the column in long_form which contains the alternative ID
        values for each row of `long_form`.
    choice_col : str, optional.
        Denotes the column in long_form which contains a one if the alternative
        pertaining to the given row was the observed outcome for the
        observation pertaining to the given row and a zero otherwise.
        Default == None.
    nest_spec : OrderedDict, or None, optional.
        Keys are strings that define the name of the nests. Values are lists of
        alternative ids, denoting which alternatives belong to which nests.
        Each alternative id must only be associated with a single nest!
        Default == None.
    mix_id_col : str, optional.
        Denotes the column in long_form that contains the identification values
        used to denote the units of observation over which parameters are
        randomly distributed.
    dense : bool, optional.
        Determines whether or not scipy sparse matrices will be returned or
        dense numpy arrays.

    Returns
    -------
    mapping_dict : OrderedDict.
        Keys will be `["rows_to_obs", "rows_to_alts", "chosen_row_to_obs",
        "rows_to_nests"]`. If `choice_col` is None, then the value for
        `chosen_row_to_obs` will be None. Likewise, if `nest_spec` is None,
        then the value for `rows_to_nests` will be None. The value for
        "rows_to_obs" will map the rows of the `long_form` to the unique
        observations (on the columns) in their order of appearance. The value
        for `rows_to_alts` will map the rows of the `long_form` to the unique
        alternatives which are possible in the dataset (on the columns), in
        sorted order--not order of appearance. The value for
        `chosen_row_to_obs`, if not None, will map the rows of the `long_form`
        that contain the chosen alternatives to the specific observations those
        rows are associated with (denoted by the columns). The value of
        `rows_to_nests`, if not None, will map the rows of the `long_form` to
        the nest (denoted by the column) that contains the row's alternative.
        If `dense==True`, the returned values will be dense numpy arrays.
        Otherwise, the returned values will be scipy sparse arrays.
    """
    # Get the id_values from the long_form dataframe
    obs_id_values = long_form[obs_id_col].values
    alt_id_values = long_form[alt_id_col].values

    # Create a matrix with the same number of rows as long_form but a single
    # column for each of the unique IDs. This matrix will associate each row
    # as belonging to a particular observation using a one and using a zero to
    # show non-association.
    rows_to_obs = create_sparse_mapping(obs_id_values)

    # Determine all of the unique alternative IDs
    all_alternatives = np.sort(np.unique(alt_id_values))

    # Create a matrix with the same number of rows as long_form but a single
    # column for each of the unique alternatives. This matrix will associate
    # each row as belonging to a particular alternative using a one and using
    # a zero to show non-association.
    rows_to_alts = create_sparse_mapping(alt_id_values,
                                         unique_ids=all_alternatives)

    if choice_col is not None:
        # Create a matrix to associate each row with the same number of
        # rows as long_form but a 1 only if that row corresponds to an
        # alternative that a given observation (denoted by the columns)
        # chose.
        chosen_row_to_obs = csr_matrix(rows_to_obs.multiply(
                                long_form[choice_col].values[:, None]))
    else:
        chosen_row_to_obs = None

    if nest_spec is not None:
        # Determine how many nests there are
        num_nests = len(nest_spec)
        # Create a mapping between the alternative ids and their nests
        alt_id_to_nest_name = {}
        for key in nest_spec:
            for element in nest_spec[key]:
                alt_id_to_nest_name[element] = key

        # Create a mapping between nest names and nest ids
        nest_ids = np.arange(1, num_nests + 1)
        nest_name_to_nest_id = dict(zip(nest_spec.keys(), nest_ids))

        # Create an array of the nest ids of each row
        nest_id_vec = np.array([nest_name_to_nest_id[alt_id_to_nest_name[x]]
                                for x in alt_id_values])

        # Create the mapping matrix between each row and the nests
        rows_to_nests = create_sparse_mapping(nest_id_vec, unique_ids=nest_ids)

    else:
        rows_to_nests = None

    if mix_id_col is not None:
        # Create a mapping matrix between each row and each 'mixing unit'
        mix_id_array = long_form[mix_id_col].values
        rows_to_mixers = create_sparse_mapping(mix_id_array)

    else:
        rows_to_mixers = None

    # Create the dictionary of mapping matrices that is to be returned
    mapping_dict = OrderedDict()
    mapping_dict["rows_to_obs"] = rows_to_obs
    mapping_dict["rows_to_alts"] = rows_to_alts
    mapping_dict["chosen_row_to_obs"] = chosen_row_to_obs
    mapping_dict["rows_to_nests"] = rows_to_nests
    mapping_dict["rows_to_mixers"] = rows_to_mixers

    # Return the dictionary of mapping matrices.
    # If desired, convert the mapping matrices to dense matrices
    if dense:
        for key in mapping_dict:
            if mapping_dict[key] is not None:
                mapping_dict[key] = mapping_dict[key].A

    return mapping_dict


def convert_long_to_wide(long_data,
                         ind_vars,
                         alt_specific_vars,
                         subset_specific_vars,
                         obs_id_col,
                         alt_id_col,
                         choice_col,
                         alt_name_dict=None,
                         null_value=np.nan):
    """
    Converts a 'long format' dataframe of cross-sectional discrete choice data
    into a 'wide format' version of the same data.

    Parameters
    ----------
    long_data : pandas dataframe.
        Contains one row for each available alternative for each observation.
        Should have the specified `[obs_id_col, alt_id_col, choice_col]` column
        headings. The dtypes of all columns should be numeric.
    ind_vars : list of strings.
        Each element should be a column heading in `long_data` that denotes a
        variable that varies across observations but not across alternatives.
    alt_specific_vars : list of strings.
        Each element should be a column heading in `long_data` that denotes a
        variable that varies not only across observations but also across all
        alternatives.
    subset_specific_vars : dict.
        Each key should be a string that is a column heading of `long_data`.
        Each value should be a list of alternative ids denoting the subset of
        alternatives which the variable (i.e. the key) over actually varies.
        These variables should vary across individuals and across some
        alternatives.
    obs_id_col : str.
        Denotes the column in `long_data` that contains the observation ID
        values for each row.
    alt_id_col : str.
        Denotes the column in `long_data` that contains the alternative ID
        values for each row.
    choice_col : str.
        Denotes the column in `long_data` that contains a one if the
        alternative pertaining to the given row was the observed outcome for
        the observation pertaining to the given row and a zero otherwise.
    alt_name_dict : dict or None, optional
        If not None, should be a dictionary whose keys are the possible values
        in `long_data[alt_id_col].unique()`. The values should be the name
        that one wants to associate with each alternative id. Default == None.
    null_value : int, float, long, or `np.nan`, optional.
        The passed value will be used to fill cells in the wide format
        dataframe when that cell is unknown for a given individual. This is
        most commonly the case when there is a variable that varies across
        alternatives and one of the alternatives is not available for a given
        indvidual. The `null_value` will be inserted for that individual for
        that variable. Default == `np.nan`.

    Returns
    -------
    final_wide_df : pandas dataframe.
        Will contain one row per observational unit. Will contain an
        observation id column of the same name as `obs_id_col`. Will also
        contain a choice column of the same name as `choice_col`. Will contain
        one availability column per unique, observed alternative in the
        dataset. Will contain one column per variable in `ind_vars`. Will
        contain one column per alternative per variable in `alt_specific_vars`.
        Will contain one column per specified alternative per variable in
        `subset_specific_vars`.
    """
    ##########
    # Check that all columns of long_data are being
    # used in the conversion to wide format
    ##########
    num_vars_accounted_for = sum([len(x) for x in
                                  [ind_vars, alt_specific_vars,
                                   subset_specific_vars,
                                   [obs_id_col, alt_id_col, choice_col]]])

    ensure_all_columns_are_used(num_vars_accounted_for, long_data)

    ##########
    # Check that all columns one wishes to use are actually in long_data
    ##########
    ensure_columns_are_in_dataframe(ind_vars,
                                    long_data,
                                    col_title="ind_vars",
                                    data_title='long_data')

    ensure_columns_are_in_dataframe(alt_specific_vars,
                                    long_data,
                                    col_title="alt_specific_vars",
                                    data_title='long_data')

    ensure_columns_are_in_dataframe(subset_specific_vars.keys(),
                                    long_data,
                                    col_title="subset_specific_vars",
                                    data_title='long_data')

    identifying_cols = [choice_col, obs_id_col, alt_id_col]
    identifying_col_string = "[choice_col, obs_id_col, alt_id_col]"
    ensure_columns_are_in_dataframe(identifying_cols,
                                    long_data,
                                    col_title=identifying_col_string,
                                    data_title='long_data')

    ##########
    # Make sure that each observation-alternative pair is unique
    ##########
    check_dataframe_for_duplicate_records(obs_id_col, alt_id_col, long_data)

    ##########
    # Make sure each observation chose an alternative that's available.
    ##########
    # Make sure that the number of chosen alternatives equals the number of
    # individuals.
    ensure_num_chosen_alts_equals_num_obs(obs_id_col, choice_col, long_data)

    ##########
    # Check that the alternative ids in the alt_name_dict are actually the
    # alternative ids used in the long_data alt_id column.
    ##########
    if alt_name_dict is not None:
        check_type_and_values_of_alt_name_dict(alt_name_dict,
                                               alt_id_col,
                                               long_data)

    ##########
    # Figure out how many rows/columns should be in the wide format dataframe
    ##########
    # Note that the number of rows in wide format is the number of observations
    num_obs = long_data[obs_id_col].unique().shape[0]

    # Figure out the total number of possible alternatives for the dataset
    num_alts = long_data[alt_id_col].unique().shape[0]

    ############
    # Calculate the needed number of columns
    ############
    # For each observation, there is at least one column-- the observation id,
    num_cols = 1
    # We should have one availability column per alternative in the dataset
    num_cols += num_alts
    # We should also have one column to record the choice of each observation
    num_cols += 1
    # We should also have one column for each individual specific variable
    num_cols += len(ind_vars)
    # We should also have one column for each alternative specific variable,
    # for each alternative
    num_cols += len(alt_specific_vars) * num_alts
    # We should have one column for each subset alternative specific variable
    # for each alternative over which the variable varies
    for col in subset_specific_vars:
        num_cols += len(subset_specific_vars[col])

    ##########
    # Create the columns of the new dataframe
    ##########
    #####
    # Create the individual specific variable columns,
    # along with the observation id column
    #####
    new_df = long_data[[obs_id_col] + ind_vars].drop_duplicates()

    # Reset the index so that the index is not based on long_data
    new_df.reset_index(inplace=True)

    #####
    # Create the choice column in the wide data format
    #####
    new_df[choice_col] = long_data.loc[long_data[choice_col] == 1,
                                       alt_id_col].values

    #####
    # Create the availability columns
    #####
    # Get the various long form mapping matrices
    mapping_res = create_long_form_mappings(long_data,
                                            obs_id_col,
                                            alt_id_col)
    row_to_obs = mapping_res["rows_to_obs"]
    row_to_alt = mapping_res["rows_to_alts"]

    # Get the matrix of observations (rows) to available alternatives (columns)
    obs_to_alt = row_to_obs.T.dot(row_to_alt).todense()

    # Determine the unique alternative IDs in the order used in obs_to_alt
    alt_id_values = long_data[alt_id_col].values
    all_alternatives = np.sort(np.unique(alt_id_values))

    # Create the names for the availability columns
    if alt_name_dict is None:
        availability_col_names = ["availability_{}".format(int(x))
                                  for x in all_alternatives]
    else:
        availability_col_names = ["availability_{}".format(alt_name_dict[x])
                                  for x in all_alternatives]

    # Create a dataframe containing the availability columns for this dataset
    availability_df = pd.DataFrame(obs_to_alt,
                                   columns=availability_col_names)

    #####
    # Create the alternative specific and subset
    # alternative specific variable columns
    #####
    # For each alternative specific variable, create a wide format dataframe
    alt_specific_dfs = []
    for col in alt_specific_vars + list(subset_specific_vars.keys()):
        # Get the relevant values from the long format dataframe
        relevant_vals = long_data[col].values[:, None]
        # Create an wide format array of the relevant values
        obs_to_var = row_to_obs.T.dot(row_to_alt.multiply(relevant_vals))
        # Ensure that the wide format array is an ndarray with of dtype float
        if issparse(obs_to_var):
            obs_to_var = obs_to_var.toarray()
        # Ensure that obs_to_var has a float dtype
        obs_to_var = obs_to_var.astype(float)

        # Place a null value in columns where the alternative is not available
        # to a given observation
        if (obs_to_alt == 0).any():
            obs_to_var[np.where(obs_to_alt == 0)] = null_value

        # Create column names for the alternative specific variable columns
        if alt_name_dict is None:
            obs_to_var_names = ["{}_{}".format(col, int(x))
                                for x in all_alternatives]
        else:
            obs_to_var_names = ["{}_{}".format(col, alt_name_dict[x])
                                for x in all_alternatives]

        # Subset obs_to_vars and obs_to_var_names if col is in
        # subset_specific_vars
        if col in subset_specific_vars:
            # Calculate the relevant column indices for
            # the specified subset of alternatives
            relevant_alt_ids = subset_specific_vars[col]
            relevant_col_idx = np.where(np.in1d(all_alternatives,
                                                relevant_alt_ids))[0]
        else:
            relevant_col_idx = None

        # Create a dataframe containing the alternative specific variables
        # or the subset alternative specific variables for the given
        # variable in the long format dataframe
        if relevant_col_idx is None:
            obs_to_var_df = pd.DataFrame(obs_to_var,
                                         columns=obs_to_var_names)
        else:
            obs_to_var_df = pd.DataFrame(obs_to_var[:, relevant_col_idx],
                                         columns=[obs_to_var_names[x] for
                                                  x in relevant_col_idx])

        # Store the current alternative specific variable columns/dataframe
        alt_specific_dfs.append(obs_to_var_df)

    # Combine all of the various alternative specific variable dataframes
    final_alt_specific_df = pd.concat(alt_specific_dfs, axis=1)

    ##########
    # Construct the final wide format dataframe to be returned
    ##########
    final_wide_df = pd.concat([new_df[[obs_id_col]],
                               new_df[[choice_col]],
                               availability_df,
                               new_df[ind_vars],
                               final_alt_specific_df],
                              axis=1)

    # Make sure one has the correct number of rows and columns in
    # the final dataframe
    if final_wide_df.shape != (num_obs, num_cols):
        msg_1 = "There is an error with the dataframe that will be returned"
        msg_2 = "The shape of the dataframe should be {}".format((num_obs,
                                                                  num_cols))
        msg_3 = "Instead, the returned dataframe will have shape: {}"
        total_msg = msg_1 + msg_2 + msg_3.format(final_wide_df.shape)
        warnings.warn(total_msg)

    # Return the wide format dataframe
    return final_wide_df


def check_wide_data_for_blank_choices(choice_col, wide_data):
    """
    Checks `wide_data` for null values in the choice column, and raises a
    helpful ValueError if null values are found.

    Parameters
    ----------
    choice_col : str.
        Denotes the column in `wide_data` that is used to record each
        observation's choice.
    wide_data : pandas dataframe.
        Contains one row for each observation. Should contain `choice_col`.

    Returns
    -------
    None.
    """
    if wide_data[choice_col].isnull().any():
        msg_1 = "One or more of the values in wide_data[choice_col] is null."
        msg_2 = " Remove null values in the choice column or fill them in."
        raise ValueError(msg_1 + msg_2)

    return None


def ensure_unique_obs_ids_in_wide_data(obs_id_col, wide_data):
    """
    Ensures that there is one observation per row in wide_data. Raises a
    helpful ValueError if otherwise.

    Parameters
    ----------
    obs_id_col : str.
        Denotes the column in `wide_data` that contains the observation ID
        values for each row.
    wide_data : pandas dataframe.
        Contains one row for each observation. Should contain the specified
        `obs_id_col` column.

    Returns
    -------
    None.
    """
    if len(wide_data[obs_id_col].unique()) != wide_data.shape[0]:
        msg = "The values in wide_data[obs_id_col] are not unique, "
        msg_2 = "but they need to be."
        raise ValueError(msg + msg_2)

    return None


def ensure_chosen_alternatives_are_in_user_alt_ids(choice_col,
                                                   wide_data,
                                                   availability_vars):
    """
    Ensures that all chosen alternatives in `wide_df` are present in the
    `availability_vars` dict. Raises a helpful ValueError if not.

    Parameters
    ----------
    choice_col : str.
        Denotes the column in `wide_data` that contains a one if the
        alternative pertaining to the given row was the observed outcome for
        the observation pertaining to the given row and a zero otherwise.
    wide_data : pandas dataframe.
        Contains one row for each observation. Should contain the specified
        `choice_col` column.
    availability_vars : dict.
        There should be one key value pair for each alternative that is
        observed in the dataset. Each key should be the alternative id for the
        alternative, and the value should be the column heading in `wide_data`
        that denotes (using ones and zeros) whether an alternative is
        available/unavailable, respectively, for a given observation.
        Alternative id's, i.e. the keys, must be integers.

    Returns
    -------
    None.
    """
    if not wide_data[choice_col].isin(availability_vars.keys()).all():
        msg = "One or more values in wide_data[choice_col] is not in the user "
        msg_2 = "provided alternative ids in availability_vars.keys()"
        raise ValueError(msg + msg_2)

    return None


def ensure_each_wide_obs_chose_an_available_alternative(obs_id_col,
                                                        choice_col,
                                                        availability_vars,
                                                        wide_data):
    """
    Checks whether or not each observation with a restricted choice set chose
    an alternative that was personally available to him or her. Will raise a
    helpful ValueError if this is not the case.

    Parameters
    ----------
    obs_id_col : str.
        Denotes the column in `wide_data` that contains the observation ID
        values for each row.
    choice_col : str.
        Denotes the column in `wide_data` that contains a one if the
        alternative pertaining to the given row was the observed outcome for
        the observation pertaining to the given row and a zero otherwise.
    availability_vars : dict.
        There should be one key value pair for each alternative that is
        observed in the dataset. Each key should be the alternative id for the
        alternative, and the value should be the column heading in `wide_data`
        that denotes (using ones and zeros) whether an alternative is
        available/unavailable, respectively, for a given observation.
        Alternative id's, i.e. the keys, must be integers.
    wide_data : pandas dataframe.
        Contains one row for each observation. Should have the specified
        `[obs_id_col, choice_col] + availability_vars.values()` columns.

    Returns
    -------
    None
    """
    # Determine the various availability values for each observation
    wide_availability_values = wide_data[list(
        availability_vars.values())].values

    # Isolate observations for whom one or more alternatives are unavailable
    unavailable_condition = ((wide_availability_values == 0).sum(axis=1)
                                                            .astype(bool))

    # Iterate over the observations with one or more unavailable alternatives
    # Check that each such observation's chosen alternative was available
    problem_obs = []
    for idx, row in wide_data.loc[unavailable_condition].iterrows():
        if row.at[availability_vars[row.at[choice_col]]] != 1:
            problem_obs.append(row.at[obs_id_col])

    if problem_obs != []:
        msg = "The following observations chose unavailable alternatives:\n{}"
        raise ValueError(msg.format(problem_obs))

    return None


def ensure_all_wide_alt_ids_are_chosen(choice_col,
                                       alt_specific_vars,
                                       availability_vars,
                                       wide_data):
    """
    Checks to make sure all user-specified alternative id's, both in
    `alt_specific_vars` and `availability_vars` are observed in the choice
    column of `wide_data`.
    """
    sorted_alt_ids = np.sort(wide_data[choice_col].unique())
    try:
        problem_ids = [x for x in availability_vars
                       if x not in sorted_alt_ids]
        problem_type = "availability_vars"
        assert problem_ids == []

        problem_ids = []
        for new_column in alt_specific_vars:
            for alt_id in alt_specific_vars[new_column]:
                if alt_id not in sorted_alt_ids and alt_id not in problem_ids:
                    problem_ids.append(alt_id)
        problem_type = "alt_specific_vars"
        assert problem_ids == []
    except AssertionError:
        msg = "The following alternative ids from {} are not "
        msg_2 = "observed in wide_data[choice_col]:\n{}"
        raise ValueError(msg.format(problem_type) + msg_2.format(problem_ids))

    return None


def ensure_contiguity_in_observation_rows(obs_id_vector):
    """
    Ensures that all rows pertaining to a given choice situation are located
    next to one another. Raises a helpful ValueError otherwise. This check is
    needed because the hessian calculation function requires the design matrix
    to have contiguity in rows with the same observation id.

    Parameters
    ----------
    rows_to_obs : 2D scipy sparse array.
        Should map each row of the long format dataferame to the unique
        observations in the dataset.
    obs_id_vector : 1D ndarray of ints.
        Should contain the id (i.e. a unique integer) that corresponds to each
        choice situation in the dataset.

    Returns
    -------
    None.
    """
    # Check that the choice situation id for each row is larger than or equal
    # to the choice situation id of the preceding row.
    contiguity_check_array = (obs_id_vector[1:] - obs_id_vector[:-1]) >= 0
    if not contiguity_check_array.all():
        problem_ids = obs_id_vector[np.where(~contiguity_check_array)]
        msg_1 = "All rows pertaining to a given choice situation must be "
        msg_2 = "contiguous. \nRows pertaining to the following observation "
        msg_3 = "id's are not contiguous: \n{}"
        raise ValueError(msg_1 + msg_2 + msg_3.format(problem_ids.tolist()))
    else:
        return None


def convert_wide_to_long(wide_data,
                         ind_vars,
                         alt_specific_vars,
                         availability_vars,
                         obs_id_col,
                         choice_col,
                         new_alt_id_name=None):
    """
    Will convert a cross-sectional dataframe of discrete choice data from wide
    format to long format.

    Parameters
    ----------
    wide_data : pandas dataframe.
        Contains one row for each observation. Should have the specified
        `[obs_id_col, choice_col] + availability_vars.values()` columns.
    ind_vars : list of strings.
        Each element should be a column heading in `wide_data` that denotes a
        variable that varies across observations but not across alternatives.
    alt_specific_vars : dict.
        Each key should be a string that will be a column heading of the
        returned, long format dataframe. Each value should be a dictionary
        where the inner key is the alternative id and the value is the column
        heading in wide data that specifies the value of the outer key for the
        associated alternative. The variables denoted by the outer key should
        vary across individuals and across some or all alternatives.
    availability_vars : dict.
        There should be one key value pair for each alternative that is
        observed in the dataset. Each key should be the alternative id for the
        alternative, and the value should be the column heading in `wide_data`
        that denotes (using ones and zeros) whether an alternative is
        available/unavailable, respectively, for a given observation.
        Alternative id's, i.e. the keys, must be integers.
    obs_id_col : str.
        Denotes the column in `wide_data` that contains the observation ID
        values for each row.
    choice_col : str.
        Denotes the column in `wide_data` that contains a one if the
        alternative pertaining to the given row was the observed outcome for
        the observation pertaining to the given row and a zero otherwise.
    new_alt_id_name : str, optional.
        If not None, should be a string. This string will be used as the column
        heading for the alternative id column in the returned 'long' format
        dataframe. If not passed, this column will be called `'alt_id'`.
        Default == None.

    Returns
    -------
    final_long_df : pandas dataframe.
        Will contain one row for each available alternative for each
        observation. Will contain an observation id column of the same name as
        `obs_id_col`. Will also contain a choice column of the same name as
        `choice_col`. Will also contain an alternative id column called
        `alt_id` if `new_alt_id_col == None`, or `new_alt_id` otherwise. Will
        contain one column per variable in `ind_vars`. Will contain one column
        per key in `alt_specific_vars`.
    """

    # Avoid to modify the original dataframe
    wide_data = wide_data.copy()

    list_availability_vars = list(availability_vars.values())
    # Allow to specify that an alternative is available to all decision-makers
    # It creates a dummy variable to specify that all alternatives are available
    # for all individuals.
    if 'available_for_all' in list_availability_vars:
        wide_data["available_for_all"] = 1
    # Allow to specify that an alternative isn't available to any decision-maker
    if 'available_for_none' in list_availability_vars:
        wide_data["available_for_none"] = 0

    ##########
    # Check that all columns of wide_data are being
    # used in the conversion to long format
    ##########
    all_alt_specific_cols = []
    for var_dict in alt_specific_vars.values():
        all_alt_specific_cols.extend(var_dict.values())

    vars_accounted_for = set(ind_vars +
                             # conver to list explicitly to support
                             # both python 2 and 3
                             list_availability_vars +
                             [obs_id_col, choice_col] +
                             all_alt_specific_cols)
    num_vars_accounted_for = len(vars_accounted_for)

    ensure_all_columns_are_used(num_vars_accounted_for,
                                wide_data,
                                data_title='wide_data')

    ##########
    # Check that all columns one wishes to use are actually in wide_data
    ##########
    ensure_columns_are_in_dataframe(ind_vars,
                                    wide_data,
                                    col_title='ind_vars',
                                    data_title='wide_data')

    ensure_columns_are_in_dataframe(availability_vars.values(),
                                    wide_data,
                                    col_title='availability_vars',
                                    data_title='wide_data')

    for new_column in alt_specific_vars:
        for alt_id in alt_specific_vars[new_column]:
            old_column = alt_specific_vars[new_column][alt_id]
            ensure_columns_are_in_dataframe([old_column],
                                            wide_data,
                                            col_title="alt_specific_vars",
                                            data_title='wide_data')

    ensure_columns_are_in_dataframe([choice_col, obs_id_col],
                                    wide_data,
                                    col_title='[choice_col, obs_id_col]',
                                    data_title='wide_data')

    ##########
    # Check the integrity of the various columns present in wide_data
    ##########
    # Make sure the observation id's are unique (i.e. one per row)
    ensure_unique_obs_ids_in_wide_data(obs_id_col, wide_data)

    # Make sure there are no blank values in the choice column
    check_wide_data_for_blank_choices(choice_col, wide_data)

    ##########
    # Check that the user-provided alternative ids are observed
    # in the realized choices.
    ##########
    ensure_all_wide_alt_ids_are_chosen(choice_col,
                                       alt_specific_vars,
                                       availability_vars,
                                       wide_data)

    ##########
    # Check that the realized choices are all in the
    # user-provided alternative ids
    ##########
    ensure_chosen_alternatives_are_in_user_alt_ids(choice_col,
                                                   wide_data,
                                                   availability_vars)

    ##########
    # Make sure each observation chose a personally available alternative.
    ##########
    ensure_each_wide_obs_chose_an_available_alternative(obs_id_col,
                                                        choice_col,
                                                        availability_vars,
                                                        wide_data)

    ##########
    # Figure out how many rows/columns should be in the long format dataframe
    ##########
    # Note that the number of rows in long format is the
    # number of available alternatives across all observations
    sorted_alt_ids = np.sort(wide_data[choice_col].unique())
    sorted_availability_cols = [availability_vars[x] for x in sorted_alt_ids]
    num_rows = wide_data[sorted_availability_cols].sum(axis=0).sum()

    #####
    # Calculate the needed number of columns
    #####
    # For each observation, there is at least one column-- the observation id,
    num_cols = 1
    # We should also have one alternative id column
    num_cols += 1
    # We should also have one column to record the choice of each observation
    num_cols += 1
    # We should also have one column for each individual specific variable
    num_cols += len(ind_vars)
    # We should also have one column for each alternative specific variable,
    num_cols += len(alt_specific_vars.keys())

    ##########
    # Create the columns of the new dataframe
    ##########
    #####
    # Create the observation id column,
    #####
    # Determine the various availability values for each observation
    wide_availability_values = wide_data[list(
        availability_vars.values())].values
    new_obs_id_col = (wide_availability_values *
                      wide_data[obs_id_col].values[:, None]).ravel()
    # Make sure the observation id column has an integer data type
    new_obs_id_col = new_obs_id_col.astype(int)

    #####
    # Create the independent variable columns. Store them in a list.
    #####
    new_ind_var_cols = []
    for var in ind_vars:
        new_ind_var_cols.append((wide_availability_values *
                                 wide_data[var].values[:, None]).ravel())

    #####
    # Create the choice column in the long data format
    #####
    wide_choice_data = (wide_data[choice_col].values[:, None] ==
                        sorted_alt_ids[None, :])
    new_choice_col = wide_choice_data.ravel()
    # Make sure the choice column has an integer data type
    new_choice_col = new_choice_col.astype(int)

    #####
    # Create the alternative id column
    #####
    new_alt_id_col = (wide_availability_values *
                      sorted_alt_ids[None, :]).ravel().astype(int)
    # Make sure the alternative id column has an integer data type
    new_alt_id_col = new_alt_id_col.astype(int)

    #####
    # Create the alternative specific and subset
    # alternative specific variable columns
    #####
    # For each alternative specific variable, create a wide format array.
    # Then unravel that array to have the long format column for the
    # alternative specific variable. Store all the long format columns
    # in a list
    new_alt_specific_cols = []
    for new_col in alt_specific_vars:
        new_wide_alt_specific_cols = []
        for alt_id in sorted_alt_ids:
            # This will extract the correct values for the alternatives over
            # which the alternative specific variables vary
            if alt_id in alt_specific_vars[new_col]:
                rel_wide_column = alt_specific_vars[new_col][alt_id]
                new_col_vals = wide_data[rel_wide_column].values[:, None]
                new_wide_alt_specific_cols.append(new_col_vals)
            # This will create placeholder zeros for the alternatives that
            # the alternative specific variables do not vary over
            else:
                new_wide_alt_specific_cols.append(np.zeros(
                                                    (wide_data.shape[0], 1)))
        concatenated_long_column = np.concatenate(new_wide_alt_specific_cols,
                                                  axis=1).ravel()
        new_alt_specific_cols.append(concatenated_long_column)

    ##########
    # Construct the final wide format dataframe to be returned
    ##########
    # Identify rows that correspond to unavailable alternatives
    availability_condition = wide_availability_values.ravel() != 0

    # Figure out the names of all of the columns in the final
    # dataframe
    alt_id_column_name = ("alt_id" if new_alt_id_name is None
                          else new_alt_id_name)
    final_long_columns = ([obs_id_col,
                           alt_id_column_name,
                           choice_col] +
                          ind_vars +
                          list(alt_specific_vars.keys()))

    # Create a 'record array' of the final dataframe's columns
    # Note that record arrays are constructed from a list of 1D
    # arrays hence the array unpacking performed below for
    # new_ind_var_cols and new_alt_specific_cols
    all_arrays = ([new_obs_id_col,
                   new_alt_id_col,
                   new_choice_col] +
                  new_ind_var_cols +
                  new_alt_specific_cols)

    # Be sure to remove rows corresponding to unavailable alternatives
    # When creating the record array.
    df_recs = np.rec.fromarrays([all_arrays[pos][availability_condition]
                                 for pos in range(len(all_arrays))],
                                names=final_long_columns)

    # Create the final dataframe
    final_long_df = pd.DataFrame.from_records(df_recs)

    ##########
    # Make sure one has the correct number of rows and columns in
    # the final dataframe
    ##########
    try:
        assert final_long_df.shape == (num_rows, num_cols)
    except AssertionError:
        msg_1 = "There is an error with the dataframe that will be returned."
        msg_2 = "The shape of the dataframe should be {}".format((num_rows,
                                                                  num_cols))
        msg_3 = "Instead, the returned dataframe will have shape: {}"
        total_msg = "\n".join([msg_1, msg_2, msg_3])
        warnings.warn(total_msg.format(final_long_df.shape))

    # Return the wide format dataframe
    return final_long_df


def convert_mixing_names_to_positions(mixing_names, ind_var_names):
    """
    Parameters
    ----------
    mixing_names : list.
        All elements should be strings. Denotes the names of the index
        variables that are being treated as random variables.
    ind_var_names : list.
        All elements should be strings, representing (in order) the variables
        in the index.

    Returns
    -------
     list.
         All elements should be ints. Elements will be the position of each of
         the elements in mixing name, in the `ind_var_names` list.
    """
    return [ind_var_names.index(name) for name in mixing_names]


def divide_long_train_test(long_data,
                           train_size,
                           obs_id_col,
                           random_state=None):
    """
    Divides a long-format DataFrame into a train and test set.

    Parameters
    ----------
    long_data : pandas dataframe.
        Contains one row for each available alternative for each observation.
        Should have the specified `[obs_id_col, alt_id_col, choice_col]` column
        headings. The dtypes of all columns should be numeric.
    train_size : float.
        Indicates the relative size of the train set. It should be a value in
        the range (0, 1).
    obs_id_col : str.
        Denotes the column in `long_data` that contains the observation ID
        values for each row.
    random_state : int or None.
        Is the seed used by the random number generator.

    Returns
    -------
    tuple.
        Tuple of two pandas dataframe. The first ones contains the train set and
        the second one contains the test set.
    """

    if not isinstance(long_data, pd.DataFrame):
        msg = "The input variable long_data must be a Pandas DataFrame."
        raise ValueError(msg)

    if train_size <= 0 or train_size >= 1:
        msg_1 = "The train_size defined is not valid."
        msg_2 = "train_size should be a value in the range (0, 1)."
        msg_3 = "The passed train_size was: {}".format(train_size)
        total_msg = "\n".join([msg_1, msg_2, msg_3])
        raise ValueError(total_msg)

    # Obtain all the observations from the long format dataframe
    obs = long_data[obs_id_col].unique()
    n_obs = long_data[obs_id_col].nunique()

    # Calculate the number of observations for train and test dataframes
    n_train = int(np.round(n_obs * train_size, decimals=0))
    n_test = n_obs - n_train
    # Divide all the dataframe indices between train and test
    np.random.seed(random_state)
    train_obs = list(np.random.choice(obs, size=n_train, replace=False))
    test_obs = [x for x in obs if x not in train_obs]
    # Shuffle test indices (train indices are shuffled when random is applied)
    random.Random(random_state).shuffle(test_obs)

    # Generate the train and test dataframes
    train_set = long_data[long_data[obs_id_col].isin(train_obs)].copy()
    test_set = long_data[long_data[obs_id_col].isin(test_obs)].copy()

    return (train_set, test_set)


def KFold_cross_val(long_data,
                    obs_id_col,
                    n_splits=5,
                    shuffle=False,
                    random_state=None):
    """
    K-Folds cross-validator.

    Provides train/test indices to split data in train/test sets. Split
    dataset into k consecutive folds (without shuffling by default).

    Parameters
    ----------
    long_data : pandas dataframe.
        Contains one row for each available alternative for each observation.
        Should have the specified `[obs_id_col, alt_id_col, choice_col]` column
        headings. The dtypes of all columns should be numeric.
    obs_id_col : str.
        Denotes the column in `data` that contains the observation ID
        values for each row.
    n_splits : int, optional.
        Number of folds. Must be at least 2. Default == 5.
    shuffle : boolean, optional.
        Whether to shuffle the data before splitting it into batches.
        Default == False.
    random_state : int or None, optional.
        If int, random_state is the seed used by the random number generator.
        Default == None.

    Yields
    -------
    train_set : pandas dataframe.
        The training long format pandas dataframe for that split.
    test_set : pandas dataframe.
        The testing long format pandas dataframe for that split.
    """
    if not isinstance(n_splits, int):
        msg_1 = "The number of folds must be of Integral type."
        msg_2 = "%s of type %s was passed."  % (n_splits, type(n_splits))
        total_msg = " ".join([msg_1, msg_2])
        raise ValueError(total_msg)

    if n_splits <= 1:
        msg_1 = "Error! k-fold cross-validation requires at least one"
        msg_2 = "train/test split by setting n_splits=2 or more,"
        msg_3 = "got n_splits={0}.".format(n_splits)
        total_msg = " ".join([msg_1, msg_2, msg_3])
        raise ValueError(total_msg)

    # Obtain all the observations from the long format dataframe
    obs = long_data[obs_id_col].unique()
    n_obs = long_data[obs_id_col].nunique()

    if n_splits > n_obs:
        msg_1 = "Cannot have number of splits n_splits={0}".format(n_splits)
        msg_2 = "greater than the number of different"
        msg_3 = "observations: n_observations={0}".format(n_obs)
        total_msg = " ".join([msg_1, msg_2, msg_3])
        raise ValueError(total_msg)

    indices = np.arange(n_obs)
    if shuffle:
        random.Random(random_state).shuffle(indices)

    # Calculate the size for each fold
    fold_sizes = np.full(n_splits, n_obs // n_splits, dtype=np.int)
    fold_sizes[:n_obs % n_splits] += 1

    current = 0
    for fold_size in fold_sizes:
        start, stop = current, current + fold_size
        fold_obs = indices[start:stop]
        current = stop

        test_mask = np.zeros(n_obs, dtype=np.bool)
        test_mask[fold_obs] = True

        train_indices = indices[np.logical_not(test_mask)]
        test_indices = indices[test_mask]

        train_obs = obs[train_indices]
        test_obs = obs[test_indices]

        train_set = long_data[long_data[obs_id_col].isin(train_obs)].copy()
        test_set = long_data[long_data[obs_id_col].isin(test_obs)].copy()

        yield train_set, test_set



"""
@author:    Timothy Brathwaite
@name:      Bootstrap Controller
@summary:   This module provides functions that will control the bootstrapping
            procedure.
"""
import sys
import time
import itertools
from copy import deepcopy

import numpy as np
import pandas as pd
# Import the following module for progressbar displays
from tqdm import tqdm, tqdm_notebook

from .display_names import model_type_to_display_name
from . import bootstrap_sampler as bs
from . import bootstrap_calcs as bc
from . import bootstrap_abc as abc
from .bootstrap_mle import retrieve_point_est
from .bootstrap_utils import ensure_samples_is_ndim_ndarray
from .construct_estimator import create_estimation_obj

try:
    # Python 3.x does not natively support xrange
    from past.builtins import xrange
except ImportError:
    pass


def is_kernel():
    """
    Determines whether or not one's code is executed inside of an ipython
    notebook environment.
    """
    if any([x in sys.modules for x in ['ipykernel', 'IPython']]):
        return True
    else:
        return False

# Create a progressbar iterable based on wehther we are in ipython or not.
PROGRESS = tqdm_notebook if is_kernel() else tqdm


def get_param_names(model_obj):
    """
    Extracts all the names to be displayed for the estimated parameters.

    Parameters
    ----------
    model_obj : an instance of an MNDC object.
        Should have the following attributes:
        `['ind_var_names', 'intercept_names', 'shape_names', 'nest_names']`.

    Returns
    -------
    all_names : list of strings.
        There will be one element for each estimated parameter. The order of
        the parameter names will be
        `['nest_parameters', 'shape_parameters', 'outside_intercepts',
          'index_coefficients']`.
    """
    # Get the index coefficient names
    all_names = deepcopy(model_obj.ind_var_names)
    # Add the intercept names if any exist
    if model_obj.intercept_names is not None:
        all_names = model_obj.intercept_names + all_names
    # Add the shape names if any exist
    if model_obj.shape_names is not None:
        all_names = model_obj.shape_names + all_names
    # Add the nest names if any exist
    if model_obj.nest_names is not None:
        all_names = model_obj.nest_names + all_names
    return all_names


def get_param_list_for_prediction(model_obj, replicates):
    """
    Create the `param_list` argument for use with `model_obj.predict`.

    Parameters
    ----------
    model_obj : an instance of an MNDC object.
        Should have the following attributes:
        `['ind_var_names', 'intercept_names', 'shape_names', 'nest_names']`.
        This model should have already undergone a complete estimation process.
        I.e. its `fit_mle` method should have been called without
        `just_point=True`.
    replicates : 2D ndarray.
        Should represent the set of parameter values that we now wish to
        partition for use with the `model_obj.predict` method.

    Returns
    -------
    param_list : list.
        Contains four elements, each being a numpy array. Either all of the
        arrays should be 1D or all of the arrays should be 2D. If 2D, the
        arrays should have the same number of columns. Each column being a
        particular set of parameter values that one wants to predict with.
        The first element in the list should be the index coefficients. The
        second element should contain the 'outside' intercept parameters if
        there are any, or None otherwise. The third element should contain
        the shape parameters if there are any or None otherwise. The fourth
        element should contain the nest coefficients if there are any or
        None otherwise. Default == None.
    """
    # Check the validity of the passed arguments
    ensure_samples_is_ndim_ndarray(replicates, ndim=2, name='replicates')
    # Determine the number of index coefficients, outside intercepts,
    # shape parameters, and nest parameters
    num_idx_coefs = len(model_obj.ind_var_names)

    intercept_names = model_obj.intercept_names
    num_outside_intercepts =\
        0 if intercept_names is None else len(intercept_names)

    shape_names = model_obj.shape_names
    num_shapes = 0 if shape_names is None else len(shape_names)

    nest_names = model_obj.nest_names
    num_nests = 0 if nest_names is None else len(nest_names)

    parameter_numbers =\
        [num_nests, num_shapes, num_outside_intercepts, num_idx_coefs]
    current_idx = 0
    param_list = []
    for param_num in parameter_numbers:
        if param_num == 0:
            param_list.insert(0, None)
            continue
        upper_idx = current_idx + param_num
        param_list.insert(0, replicates[:, current_idx:upper_idx].T)
        current_idx += param_num
    return param_list


def ensure_replicates_kwarg_validity(replicate_kwarg):
    """
    Ensures `replicate_kwarg` is either 'bootstrap' or 'jackknife'. Raises a
    helpful ValueError otherwise.
    """
    if replicate_kwarg not in ['bootstrap', 'jackknife']:
        msg = "`replicates` MUST be either 'bootstrap' or 'jackknife'."
        raise ValueError(msg)
    return None


class Boot(object):
    """
    Class to perform bootstrap resampling and to store and display its results.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    """
    def __init__(self,
                 model_obj,
                 mle_params):
        # Store the model object.
        self.model_obj = model_obj

        # Determine the parameter names
        param_names = get_param_names(model_obj)

        # Store the MLE parameters
        self.mle_params = pd.Series(mle_params, index=param_names)

        # Initialize the attributes that will be used later on.
        desired_attributes =\
            ["bootstrap_replicates", "jackknife_replicates",
             "percentile_interval", "bca_interval",
             "abc_interval", "all_intervals",
             "jackknife_log_likehoods",
             "bootstrap_log_likelihoods"]
        for attr_name in desired_attributes:
            setattr(self, attr_name, None)

        return None

    def generate_bootstrap_replicates(self,
                                      num_samples,
                                      mnl_obj=None,
                                      mnl_init_vals=None,
                                      mnl_fit_kwargs=None,
                                      extract_init_vals=None,
                                      print_res=False,
                                      method="BFGS",
                                      loss_tol=1e-06,
                                      gradient_tol=1e-06,
                                      maxiter=1000,
                                      ridge=None,
                                      constrained_pos=None,
                                      boot_seed=None,
                                      weights=None):
        """
        Generates the bootstrap replicates for one's given model and dataset.

        Parameters
        ----------
        num_samples : positive int.
            Specifies the number of bootstrap samples that are to be drawn.
        mnl_obj : an instance of pylogit.MNL or None, optional.
            Should be the MNL model object that is used to provide starting
            values for the final model being estimated. If None, then one's
            final model should be an MNL model. Default == None.
        mnl_init_vals : 1D ndarray or None, optional.
            If the model that is being estimated is not an MNL, then
            `mnl_init_val` should be passed. Should contain the values used to
            begin the estimation process for the MNL model that is used to
            provide starting values for our desired model. Default == None.
        mnl_fit_kwargs : dict or None.
            If the model that is being estimated is not an MNL, then
            `mnl_fit_kwargs` should be passed.
        extract_init_vals : callable or None, optional.
            Should accept 3 arguments, in the following order. First, it should
            accept `orig_model_obj`. Second, it should accept a pandas Series
            of estimated parameters from the MNL model. The Series' index will
            be the names of the coefficients from `mnl_names`. Thirdly, it
            should accept an int denoting the number of parameters in the final
            choice model. The callable should return a 1D ndarray of starting
            values for the final choice model. Default == None.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
            Default `== True`.
        method : str, optional.
            Should be a valid string for scipy.optimize.minimize. Determines
            the optimization algorithm that is used for this problem.
            Default `== 'bfgs'`.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next that is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default `== 1e-06`.
        maxiter : int, optional.
            Determines the maximum number of iterations used by the optimizer.
            Default `== 1000`.
        ridge : int, float, long, or None, optional.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero. Default `== None`.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_vals.size.` Default == None.
        boot_seed = non-negative int or None, optional.
            Denotes the random seed to be used when generating the bootstrap
            samples. If None, the sample generation process will generally be
            non-reproducible. Default == None.
        weights : 1D ndarray or None, optional.
            Allows for the calculation of weighted log-likelihoods. The weights
            can represent various things. In stratified samples, the weights
            may be the proportion of the observations in a given strata for a
            sample in relation to the proportion of observations in that strata
            in the population. In latent class models, the weights may be the
            probability of being a particular class.

        Returns
        -------
        None. Will store the bootstrap replicates on the
        `self.bootstrap_replicates` attribute.
        """
        print("Generating Bootstrap Replicates")
        print(time.strftime("%a %m-%d-%Y %I:%M%p"))
        sys.stdout.flush()
        # Check the passed arguments for validity.

        # Create an array of the observation ids
        obs_id_array = self.model_obj.data[self.model_obj.obs_id_col].values
        # Alias the alternative IDs and the Choice Array
        alt_id_array = self.model_obj.alt_IDs
        choice_array = self.model_obj.choices

        # Determine how many parameters are being estimated.
        num_params = self.mle_params.shape[0]

        # Figure out which observations are in each bootstrap sample.
        obs_id_per_sample =\
            bs.create_cross_sectional_bootstrap_samples(obs_id_array,
                                                        alt_id_array,
                                                        choice_array,
                                                        num_samples,
                                                        seed=boot_seed)

        # Get the dictionary of sub-dataframes for each observation id
        dfs_by_obs_id =\
            bs.create_deepcopied_groupby_dict(self.model_obj.data,
                                              self.model_obj.obs_id_col)

        # Create a column name for the bootstrap id columns.
        boot_id_col = "bootstrap_id"

        # Initialize an array to store the bootstrapped point estimates.
        point_estimates = np.empty((num_samples, num_params), dtype=float)

        # Get keyword arguments for final model estimation with new data.
        fit_kwargs = {"print_res": print_res,
                      "method": method,
                      "loss_tol": loss_tol,
                      "gradient_tol": gradient_tol,
                      "maxiter": maxiter,
                      "ridge": ridge,
                      "constrained_pos": constrained_pos,
                      "just_point": True}

        # Get the specification and name dictionary of the MNL model.
        mnl_spec = None if mnl_obj is None else mnl_obj.specification
        mnl_names = None if mnl_obj is None else mnl_obj.name_spec

        # Create an iterable for iteration
        iterable_for_iteration = PROGRESS(xrange(num_samples),
                                          desc="Creating Bootstrap Replicates",
                                          total=num_samples)

        # Iterate through the bootstrap samples and perform the MLE
        for row in iterable_for_iteration:
            # Get the bootstrapped dataframe
            bootstrap_df =\
                bs.create_bootstrap_dataframe(self.model_obj.data,
                                              self.model_obj.obs_id_col,
                                              obs_id_per_sample[row, :],
                                              dfs_by_obs_id,
                                              boot_id_col=boot_id_col)

            # Go through the necessary estimation routine to bootstrap the MLE.
            current_results =\
                retrieve_point_est(self.model_obj,
                                   bootstrap_df,
                                   boot_id_col,
                                   num_params,
                                   mnl_spec,
                                   mnl_names,
                                   mnl_init_vals,
                                   mnl_fit_kwargs,
                                   extract_init_vals=extract_init_vals,
                                   **fit_kwargs)

            # Store the bootstrapped point estimate.
            point_estimates[row] = current_results["x"]

        # Store the point estimates as a pandas dataframe
        self.bootstrap_replicates =\
            pd.DataFrame(point_estimates, columns=self.mle_params.index)

        # Print a 'finished' message for users
        print("Finished Generating Bootstrap Replicates")
        print(time.strftime("%a %m-%d-%Y %I:%M%p"))
        return None

    def generate_jackknife_replicates(self,
                                      mnl_obj=None,
                                      mnl_init_vals=None,
                                      mnl_fit_kwargs=None,
                                      extract_init_vals=None,
                                      print_res=False,
                                      method="BFGS",
                                      loss_tol=1e-06,
                                      gradient_tol=1e-06,
                                      maxiter=1000,
                                      ridge=None,
                                      constrained_pos=None):
        """
        Generates the jackknife replicates for one's given model and dataset.

        Parameters
        ----------
        mnl_obj : an instance of pylogit.MNL or None, optional.
            Should be the MNL model object that is used to provide starting
            values for the final model being estimated. If None, then one's
            final model should be an MNL model. Default == None.
        mnl_init_vals : 1D ndarray or None, optional.
            If the model that is being estimated is not an MNL, then
            `mnl_init_val` should be passed. Should contain the values used to
            begin the estimation process for the MNL model that is used to
            provide starting values for our desired model. Default == None.
        mnl_fit_kwargs : dict or None.
            If the model that is being estimated is not an MNL, then
            `mnl_fit_kwargs` should be passed.
        extract_init_vals : callable or None, optional.
            Should accept 3 arguments, in the following order. First, it should
            accept `orig_model_obj`. Second, it should accept a pandas Series
            of estimated parameters from the MNL model. The Series' index will
            be the names of the coefficients from `mnl_names`. Thirdly, it
            should accept an int denoting the number of parameters in the final
            choice model. The callable should return a 1D ndarray of starting
            values for the final choice model. Default == None.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
            Default `== True`.
        method : str, optional.
            Should be a valid string for scipy.optimize.minimize. Determines
            the optimization algorithm that is used for this problem.
            Default `== 'bfgs'`.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next that is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default `== 1e-06`.
        maxiter : int, optional.
            Determines the maximum number of iterations used by the optimizer.
            Default `== 1000`.
        ridge : int, float, long, or None, optional.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero. Default `== None`.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_vals.size.` Default == None.

        Returns
        -------
        None. Will store the bootstrap replicates on the
        `self.bootstrap_replicates` attribute.
        """
        print("Generating Jackknife Replicates")
        print(time.strftime("%a %m-%d-%Y %I:%M%p"))
        sys.stdout.flush()
        # Take note of the observation id column that is to be used
        obs_id_col = self.model_obj.obs_id_col

        # Get the array of original observation ids
        orig_obs_id_array =\
            self.model_obj.data[obs_id_col].values

        # Get an array of the unique observation ids.
        unique_obs_ids = np.sort(np.unique(orig_obs_id_array))

        # Determine how many observations are in one's dataset.
        num_obs = unique_obs_ids.size
        # Determine how many parameters are being estimated.
        num_params = self.mle_params.size

        # Get keyword arguments for final model estimation with new data.
        fit_kwargs = {"print_res": print_res,
                      "method": method,
                      "loss_tol": loss_tol,
                      "gradient_tol": gradient_tol,
                      "maxiter": maxiter,
                      "ridge": ridge,
                      "constrained_pos": constrained_pos,
                      "just_point": True}

        # Get the specification and name dictionary of the MNL model.
        mnl_spec = None if mnl_obj is None else mnl_obj.specification
        mnl_names = None if mnl_obj is None else mnl_obj.name_spec

        # Initialize the array of jackknife replicates
        point_replicates = np.empty((num_obs, num_params), dtype=float)

        # Create an iterable for iteration
        iterable_for_iteration = PROGRESS(enumerate(unique_obs_ids),
                                          desc="Creating Jackknife Replicates",
                                          total=unique_obs_ids.size)

        # Populate the array of jackknife replicates
        for pos, obs_id in iterable_for_iteration:
            # Create the dataframe without the current observation
            new_df = self.model_obj.data.loc[orig_obs_id_array != obs_id]
            # Get the point estimate for this new dataset
            current_results =\
                retrieve_point_est(self.model_obj,
                                   new_df,
                                   obs_id_col,
                                   num_params,
                                   mnl_spec,
                                   mnl_names,
                                   mnl_init_vals,
                                   mnl_fit_kwargs,
                                   extract_init_vals=extract_init_vals,
                                   **fit_kwargs)
            # Store the estimated parameters
            point_replicates[pos] = current_results['x']

        # Store the jackknife replicates as a pandas dataframe
        self.jackknife_replicates =\
            pd.DataFrame(point_replicates, columns=self.mle_params.index)
        # Print a 'finished' message for users
        print("Finished Generating Jackknife Replicates")
        print(time.strftime("%a %m-%d-%Y %I:%M%p"))
        return None

    def calc_log_likes_for_replicates(self,
                                      replicates='bootstrap',
                                      num_draws=None,
                                      seed=None):
        """
        Calculate the log-likelihood value of one's replicates, given one's
        dataset.

        Parameters
        ----------
        replicates : str in {'bootstrap', 'jackknife'}.
            Denotes which set of replicates should have their log-likelihoods
            calculated.
        num_draws : int greater than zero or None, optional.
            Denotes the number of random draws for mixed logit estimation. If
            None, then no random draws will be made. Default == None.
        seed : int greater than zero or None, optional.
            Denotes the random seed to be used for mixed logit estimation.
            If None, then no random seed will be set. Default == None.

        Returns
        -------
        log_likelihoods : 1D ndarray.
            Each element stores the log-likelihood of the associated parameter
            values on the model object's dataset. The log-likelihoods are also
            stored on the `replicates + '_log_likelihoods'` attribute.
        """
        # Check the validity of the kwargs
        ensure_replicates_kwarg_validity(replicates)

        # Get the desired type of replicates
        replicate_vec = getattr(self, replicates + "_replicates").values

        # Determine the choice column
        choice_col = self.model_obj.choice_col

        # Split the control flow based on whether we're using a Nested Logit
        current_model_type = self.model_obj.model_type
        non_2d_predictions =\
            [model_type_to_display_name["Nested Logit"],
             model_type_to_display_name["Mixed Logit"]]
        if current_model_type not in non_2d_predictions:
            # Get the param list for this set of replicates
            param_list =\
                get_param_list_for_prediction(self.model_obj, replicate_vec)

            # Get the 'chosen_probs' using the desired set of replicates
            chosen_probs =\
                self.model_obj.predict(self.model_obj.data,
                                       param_list=param_list,
                                       return_long_probs=False,
                                       choice_col=choice_col)
        else:
            # Initialize a list of chosen probs
            chosen_probs_list = []

            # Create an iterable for iteration
            iterable_for_iteration = PROGRESS(xrange(replicate_vec.shape[0]),
                                              desc="Calculate Gradient Norms",
                                              total=replicate_vec.shape[0])

            # Populate the list of chosen probabilities for each vector of
            # parameter values
            for idx in iterable_for_iteration:
                # Get the param list for this set of replicates
                param_list =\
                    get_param_list_for_prediction(self.model_obj,
                                                  replicate_vec[idx][None, :])
                # Use 1D parameters in the prediction function
                param_list =\
                    [x.ravel() if x is not None else x for x in param_list]

                # Get the 'chosen_probs' using the desired set of replicates
                chosen_probs =\
                    self.model_obj.predict(self.model_obj.data,
                                           param_list=param_list,
                                           return_long_probs=False,
                                           choice_col=choice_col,
                                           num_draws=num_draws,
                                           seed=seed)

                # store those chosen prob_results
                chosen_probs_list.append(chosen_probs[:, None])

            # Get the final array of chosen probs
            chosen_probs = np.concatenate(chosen_probs_list, axis=1)

        # Calculate the log_likelihood
        log_likelihoods = np.log(chosen_probs).sum(axis=0)

        # Store the log-likelihood values
        attribute_name = replicates + "_log_likelihoods"
        log_like_series = pd.Series(log_likelihoods, name=attribute_name)
        setattr(self, attribute_name, log_like_series)
        return log_likelihoods

    def calc_gradient_norm_for_replicates(self,
                                          replicates='bootstrap',
                                          ridge=None,
                                          constrained_pos=None,
                                          weights=None):
        """
        Calculate the Euclidean-norm of the gradient of one's replicates, given
        one's dataset.

        Parameters
        ----------
        replicates : str in {'bootstrap', 'jackknife'}.
            Denotes which set of replicates should have their log-likelihoods
            calculated.
        ridge : float or None, optional.
            Denotes the ridge penalty used when estimating the replicates, and
            to be used when calculating the gradient. If None, no ridge penalty
            is used. Default == None.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `self.mle_params` Default == None.
        weights : 1D ndarray or None, optional.
            Allows for the calculation of weighted log-likelihoods. The weights
            can represent various things. In stratified samples, the weights
            may be the proportion of the observations in a given strata for a
            sample in relation to the proportion of observations in that strata
            in the population. In latent class models, the weights may be the
            probability of being a particular class.

        Returns
        -------
        log_likelihoods : 1D ndarray.
            Each element stores the log-likelihood of the associated parameter
            values on the model object's dataset. The log-likelihoods are also
            stored on the `replicates + '_log_likelihoods'` attribute.
        """
        # Check the validity of the kwargs
        ensure_replicates_kwarg_validity(replicates)
        # Create the estimation object
        estimation_obj =\
            create_estimation_obj(self.model_obj,
                                  self.mle_params.values,
                                  ridge=ridge,
                                  constrained_pos=constrained_pos,
                                  weights=weights)
        # Prepare the estimation object to calculate the gradients
        if hasattr(estimation_obj, "set_derivatives"):
            estimation_obj.set_derivatives()
        # Get the array of parameter replicates
        replicate_array = getattr(self, replicates + "_replicates").values
        # Determine the number of replicates
        num_reps = replicate_array.shape[0]
        # Initialize an empty array to store the gradient norms
        gradient_norms = np.empty((num_reps,), dtype=float)

        # Create an iterable for iteration
        iterable_for_iteration = PROGRESS(xrange(num_reps),
                                          desc="Calculating Gradient Norms",
                                          total=num_reps)

        # Iterate through the rows of the replicates and calculate and store
        # the gradient norm for each replicated parameter vector.
        for row in iterable_for_iteration:
            current_params = replicate_array[row]
            gradient = estimation_obj.convenience_calc_gradient(current_params)
            gradient_norms[row] = np.linalg.norm(gradient)
        return gradient_norms

    def calc_percentile_interval(self, conf_percentage):
        """
        Calculates percentile bootstrap confidence intervals for one's model.

        Parameters
        ----------
        conf_percentage : scalar in the interval (0.0, 100.0).
            Denotes the confidence-level for the returned endpoints. For
            instance, to calculate a 95% confidence interval, pass `95`.

        Returns
        -------
        None. Will store the percentile intervals as `self.percentile_interval`

        Notes
        -----
        Must have all ready called `self.generate_bootstrap_replicates`.
        """
        # Get the alpha % that corresponds to the given confidence percentage.
        alpha = bc.get_alpha_from_conf_percentage(conf_percentage)
        # Create the column names for the dataframe of confidence intervals
        single_column_names =\
            ['{:.3g}%'.format(alpha / 2.0),
             '{:.3g}%'.format(100 - alpha / 2.0)]
        # Calculate the desired confidence intervals.
        conf_intervals =\
            bc.calc_percentile_interval(self.bootstrap_replicates.values,
                                        conf_percentage)
        # Store the desired confidence intervals
        self.percentile_interval =\
            pd.DataFrame(conf_intervals.T,
                         index=self.mle_params.index,
                         columns=single_column_names)
        return None

    def calc_bca_interval(self, conf_percentage):
        """
        Calculates Bias-Corrected and Accelerated (BCa) Bootstrap Confidence
        Intervals for one's model.

        Parameters
        ----------
        conf_percentage : scalar in the interval (0.0, 100.0).
            Denotes the confidence-level for the returned endpoints. For
            instance, to calculate a 95% confidence interval, pass `95`.

        Returns
        -------
        None. Will store the BCa intervals as `self.abc_interval`.

        Notes
        -----
        Must have all ready called `self.generate_bootstrap_replicates` and
        `self.generate_jackknife_replicates`.
        """
        # Get the alpha % that corresponds to the given confidence percentage.
        alpha = bc.get_alpha_from_conf_percentage(conf_percentage)
        # Create the column names for the dataframe of confidence intervals
        single_column_names =\
            ['{:.3g}%'.format(alpha / 2.0),
             '{:.3g}%'.format(100 - alpha / 2.0)]
        # Bundle the arguments needed to create the desired confidence interval
        args = [self.bootstrap_replicates.values,
                self.jackknife_replicates.values,
                self.mle_params.values,
                conf_percentage]
        # Calculate the BCa confidence intervals.
        conf_intervals = bc.calc_bca_interval(*args)
        # Store the BCa confidence intervals.
        self.bca_interval = pd.DataFrame(conf_intervals.T,
                                         index=self.mle_params.index,
                                         columns=single_column_names)
        return None

    def calc_abc_interval(self,
                          conf_percentage,
                          init_vals,
                          epsilon=0.001,
                          **fit_kwargs):
        """
        Calculates Approximate Bootstrap Confidence Intervals for one's model.

        Parameters
        ----------
        conf_percentage : scalar in the interval (0.0, 100.0).
            Denotes the confidence-level for the returned endpoints. For
            instance, to calculate a 95% confidence interval, pass `95`.
        init_vals : 1D ndarray.
            The initial values used to estimate the one's choice model.
        epsilon : positive float, optional.
            Should denote the 'very small' value being used to calculate the
            desired finite difference approximations to the various influence
            functions. Should be close to zero.
            Default == sys.float_info.epsilon.
        fit_kwargs : additional keyword arguments, optional.
            Should contain any additional kwargs used to alter the default
            behavior of `model_obj.fit_mle` and thereby enforce conformity with
            how the MLE was obtained. Will be passed directly to
            `model_obj.fit_mle`.

        Returns
        -------
        None. Will store the ABC intervals as `self.abc_interval`.
        """
        print("Calculating Approximate Bootstrap Confidence (ABC) Intervals")
        print(time.strftime("%a %m-%d-%Y %I:%M%p"))
        sys.stdout.flush()
        # Get the alpha % that corresponds to the given confidence percentage.
        alpha = bc.get_alpha_from_conf_percentage(conf_percentage)
        # Create the column names for the dataframe of confidence intervals
        single_column_names =\
            ['{:.3g}%'.format(alpha / 2.0),
             '{:.3g}%'.format(100 - alpha / 2.0)]
        # Calculate the ABC confidence intervals
        conf_intervals =\
            abc.calc_abc_interval(self.model_obj,
                                  self.mle_params.values,
                                  init_vals,
                                  conf_percentage,
                                  epsilon=epsilon,
                                  **fit_kwargs)
        # Store the ABC confidence intervals
        self.abc_interval = pd.DataFrame(conf_intervals.T,
                                         index=self.mle_params.index,
                                         columns=single_column_names)
        return None

    def calc_conf_intervals(self,
                            conf_percentage,
                            interval_type='all',
                            init_vals=None,
                            epsilon=abc.EPSILON,
                            **fit_kwargs):
        """
        Calculates percentile, bias-corrected and accelerated, and approximate
        bootstrap confidence intervals.

        Parameters
        ----------
        conf_percentage : scalar in the interval (0.0, 100.0).
            Denotes the confidence-level for the returned endpoints. For
            instance, to calculate a 95% confidence interval, pass `95`.
        interval_type : str in {'all', 'pi', 'bca', 'abc'}, optional.
            Denotes the type of confidence intervals that should be calculated.
            'all' results in all types of confidence intervals being
            calculated. 'pi' means 'percentile intervals', 'bca' means
            'bias-corrected and accelerated', and 'abc' means 'approximate
            bootstrap confidence' intervals. Default == 'all'.
        init_vals : 1D ndarray.
            The initial values used to estimate the one's choice model.
        epsilon : positive float, optional.
            Should denote the 'very small' value being used to calculate the
            desired finite difference approximations to the various influence
            functions for the 'abc' intervals. Should be close to zero.
            Default == sys.float_info.epsilon.
        fit_kwargs : additional keyword arguments, optional.
            Should contain any additional kwargs used to alter the default
            behavior of `model_obj.fit_mle` and thereby enforce conformity with
            how the MLE was obtained. Will be passed directly to
            `model_obj.fit_mle` when calculating the 'abc' intervals.

        Returns
        -------
        None. Will store the confidence intervals on their respective model
        objects: `self.percentile_interval`, `self.bca_interval`,
        `self.abc_interval`, or all of these objects.
        """
        if interval_type == 'pi':
            self.calc_percentile_interval(conf_percentage)
        elif interval_type == 'bca':
            self.calc_bca_interval(conf_percentage)
        elif interval_type == 'abc':
            self.calc_abc_interval(conf_percentage,
                                   init_vals,
                                   epsilon=epsilon,
                                   **fit_kwargs)
        elif interval_type == 'all':
            print("Calculating Percentile Confidence Intervals")
            sys.stdout.flush()
            self.calc_percentile_interval(conf_percentage)

            print("Calculating BCa Confidence Intervals")
            sys.stdout.flush()
            self.calc_bca_interval(conf_percentage)

            # Note we don't print a user message here since that is done in
            # self.calc_abc_interval().
            self.calc_abc_interval(conf_percentage,
                                   init_vals,
                                   epsilon=epsilon,
                                   **fit_kwargs)
            # Get the alpha % for the given confidence percentage.
            alpha = bc.get_alpha_from_conf_percentage(conf_percentage)
            # Get lists of the interval type names and the endpoint names
            interval_type_names = ['percentile_interval',
                                   'BCa_interval',
                                   'ABC_interval']
            endpoint_names = ['{:.3g}%'.format(alpha / 2.0),
                              '{:.3g}%'.format(100 - alpha / 2.0)]
            # Create the column names for the dataframe of confidence intervals
            multi_index_names =\
                list(itertools.product(interval_type_names, endpoint_names))
            df_column_index = pd.MultiIndex.from_tuples(multi_index_names)
            # Create the dataframe containing all confidence intervals
            self.all_intervals = pd.concat([self.percentile_interval,
                                            self.bca_interval,
                                            self.abc_interval],
                                           axis=1,
                                           ignore_index=True)
            # Store the column names for the combined confidence intervals
            self.all_intervals.columns = df_column_index
            self.all_intervals.index = self.mle_params.index
        else:
            msg =\
                "interval_type MUST be in `['pi', 'bca', 'abc', 'all']`"
            raise ValueError(msg)
        return None



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 25 07:19:49 2016

@name:      MultiNomial Logit
@author:    Timothy Brathwaite
            José Ángel Martín Baos
@summary:   Contains functions necessary for estimating multinomial logit
            models (with the help of the "base_multinomial_cm.py" file).
            Differs from version one since it works with the shape, intercept,
            index coefficient partitioning of estimated parameters as opposed
            to the shape, index coefficient partitioning scheme of version 1.
"""
from __future__ import absolute_import

import warnings
import numpy as np
from scipy.sparse import diags

from . import choice_calcs as cc
from . import base_multinomial_cm_v2 as base_mcm
from .estimation import LogitTypeEstimator
from .estimation import estimate
from .display_names import model_type_to_display_name

# Create a variable that will be printed if there is a non-fatal error
# in the MNL class construction
_msg_1 = "The Multinomial Logit Model has no shape parameters. "
_msg_2 = "shape_names and shape_ref_pos will be ignored if passed."
_shape_ignore_msg = _msg_1 + _msg_2

# Create a warning string that will be issued if ridge regression is performed.
_msg_3 = "NOTE: A penalized regression is being performed. The "
_msg_4 = "reported standard errors and robust standard errors "
_msg_5 = "***WILL BE INCORRECT***."
_ridge_warning_msg = _msg_3 + _msg_4 + _msg_5

# Alias necessary functions from the base multinomial choice model module
general_log_likelihood = cc.calc_log_likelihood
general_gradient = cc.calc_gradient
general_calc_probabilities = cc.calc_probabilities
general_hessian = cc.calc_hessian


def split_param_vec(beta,
                    rows_to_alts=None,
                    design=None,
                    return_all_types=False,
                    *args, **kwargs):
    """
    Parameters
    ----------
    beta : 1D ndarray.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    rows_to_alts : None,
        Not actually used. Included merely for consistency with other models.
    design : None.
        Not actually used. Included merely for consistency with other models.
    return_all_types : bool, optional.
        Determines whether or not a tuple of 4 elements will be returned (with
        one element for the nest, shape, intercept, and index parameters for
        this model). If False, a tuple of 3 elements will be returned, as
        described below.

    Returns
    -------
    tuple.
        `(None, None, beta)`. This function is merely for compatibility with
        the other choice model files.

    Note
    ----
    If `return_all_types == True` then the function will return a tuple of four
    objects. In order, these objects will either be None or the arrays
    representing the arrays corresponding to the nest, shape, intercept, and
    index parameters.
    """
    if return_all_types:
        return None, None, None, beta
    else:
        return None, None, beta


def _mnl_utility_transform(systematic_utilities, *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        Should contain the systematic utilities for each each available
        alternative for each observation.

    Returns
    -------
    `systematic_utilities[:, None]`
    """
    # Be sure to return a 2D array since other functions will be expecting this
    if len(systematic_utilities.shape) == 1:
        systematic_utilities = systematic_utilities[:, np.newaxis]

    return systematic_utilities


def _mnl_transform_deriv_c(*args, **kwargs):
    """
    Returns None.

    This is a place holder function since the MNL model has no shape
    parameters.
    """
    # This is a place holder function since the MNL model has no shape
    # parameters.
    return None


def _mnl_transform_deriv_alpha(*args, **kwargs):
    """
    Returns None.

    This is a place holder function since the MNL model has no intercept
    parameters outside of the index.
    """
    # This is a place holder function since the MNL model has no intercept
    # parameters outside the index.
    return None


class MNLEstimator(LogitTypeEstimator):
    """
    Estimation Object used to enforce uniformity in the estimation process
    across the various logit-type models.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `num_params` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    """
    def set_derivatives(self):
        # Pre-calculate the derivative of the transformation vector with
        # respect to the vector of systematic utilities
        dh_dv = diags(np.ones(self.design.shape[0]), 0, format='csr')

        # Create a function to calculate dh_dv which will return the
        # pre-calculated result when called
        def calc_dh_dv(*args):
            return dh_dv

        self.calc_dh_dv = calc_dh_dv
        self.calc_dh_d_alpha = _mnl_transform_deriv_alpha
        self.calc_dh_d_shape = _mnl_transform_deriv_c

    def check_length_of_initial_values(self, init_values):
        """
        Ensures that `init_values` is of the correct length. Raises a helpful
        ValueError if otherwise.

        Parameters
        ----------
        init_values : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each index coefficient, outside intercept
            parameter, and shape parameter being estimated.

        Returns
        -------
        None.
        """
        # Calculate the expected number of index parameters
        num_index_coefs = self.design.shape[1]

        if init_values.shape[0] != num_index_coefs:
            msg_1 = "The initial values are of the wrong dimension."
            msg_2 = "It should be of dimension {}"
            msg_3 = "But instead it has dimension {}"
            raise ValueError(msg_1 +
                             msg_2.format(num_index_coefs) +
                             msg_3.format(init_values.shape[0]))

        return None


class MNL(base_mcm.MNDC_Model):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV file
        containing the long format data for this choice model. Note long format
        is has one row per available alternative for each observation. If
        pandas dataframe, the dataframe should be the long format data for the
        choice model.
    alt_id_col :str.
        Should denote the column in data which contains the alternative
        identifiers for each row.
    obs_id_col : str.
        Should denote the column in data which contains the observation
        identifiers for each row.
    choice_col : str.
        Should denote the column in data which contains the ones and zeros that
        denote whether or not the given row corresponds to the chosen
        alternative for the given individual.
    specification : OrderedDict.
        Keys are a proper subset of the columns in `data`. Values are either a
        list or a single string, "all_diff" or "all_same". If a list, the
        elements should be:
            - single objects that are in the alternative ID column of `data`
            - lists of objects that are within the alternative ID column of
              `data`. For each single object in the list, a unique column will
              be created (i.e. there will be a unique coefficient for that
              variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification` values, a single column will be created for all
              the alternatives within the iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    names : OrderedDict, optional.
        Should have the same keys as `specification`. For each key:
            - if the corresponding value in `specification` is "all_same", then
              there should be a single string as the value in names.
            - if the corresponding value in `specification` is "all_diff", then
              there should be a list of strings as the value in names. There
              should be one string in the value in names for each possible
              alternative.
            - if the corresponding value in `specification` is a list, then
              there should be a list of strings as the value in names. There
              should be one string the value in names per item in the value in
              `specification`.
        Default == None.

    """
    def __init__(self,
                 data,
                 alt_id_col,
                 obs_id_col,
                 choice_col,
                 specification,
                 names=None,
                 *args, **kwargs):
        ##########
        # Print a helpful message for users who have included shape parameters
        # or shape names unneccessarily
        ##########
        for keyword in ["shape_names", "shape_ref_pos"]:
            if keyword in kwargs and kwargs[keyword] is not None:
                warnings.warn(_shape_ignore_msg)
                break

        if "intercept_ref_pos" in kwargs:
            if kwargs["intercept_ref_pos"] is not None:
                msg = "The MNL model should have all intercepts in the index."
                raise ValueError(msg)

        # Carry out the common instantiation process for all choice models
        super(MNL, self).__init__(data,
                                  alt_id_col,
                                  obs_id_col,
                                  choice_col,
                                  specification,
                                  names=names,
                                  model_type=model_type_to_display_name["MNL"])

        # Store the utility transform function
        self.utility_transform = _mnl_utility_transform

        return None

    def fit_mle(self,
                init_vals,
                print_res=True,
                method="BFGS",
                loss_tol=1e-06,
                gradient_tol=1e-06,
                maxiter=1000,
                PMLE=None,
                PMLE_lambda=0,
                constrained_pos=None,
                just_point=False,
                **kwargs):
        """
        Parameters
        ----------
        init_vals : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each utility coefficient being estimated.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
        method : str, optional.
            Should be a valid string that can be passed to
            scipy.optimize.minimize. Determines the optimization algorithm that
            is used for this problem. If 'em' is passed, a custom coded EM
            algorithm will be used. Default `== 'newton-cg'`.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next that is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
        PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
            It determines if a Penalized Maximum Likelihood Estimation should be
            executed. The value of the parameter determines the type of PMLE. The
            None value states that no PMLE is executed. Default = None.
        PMLE_lambda : int, float, long, or None, optional.
            Lambda parameter for LASSO or ridge regression. It should be an int,
            float or long and determines the penalty for the optimization.
            Default = 0.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_vals.size.` Default == None.
        just_point : bool, optional.
            Determines whether (True) or not (False) calculations that are non-
            critical for obtaining the maximum likelihood point estimate will
            be performed. If True, this function will return the results
            dictionary from scipy.optimize. Default == False.

        Returns
        -------
        None or dict.
            If `just_point` is False, None is returned and the estimation
            results are saved to the model instance. If `just_point` is True,
            then the results dictionary from scipy.optimize() is returned.
        """
        # Check integrity of passed arguments
        kwargs_to_be_ignored = ["init_shapes", "init_intercepts", "init_coefs"]
        if any([x in kwargs for x in kwargs_to_be_ignored]):
            msg = "MNL model does not use of any of the following kwargs:\n{}"
            msg_2 = "Remove such kwargs and pass a single init_vals argument"
            raise ValueError(msg.format(kwargs_to_be_ignored) + msg_2)

        if PMLE is not None:
            warnings.warn(_ridge_warning_msg)

        # Store the optimization method
        self.optimization_method = method

        # Construct the mappings from alternatives to observations and from
        # chosen alternatives to observations
        mapping_res = self.get_mappings_for_fit()

        # Create the estimation object
        zero_vector = np.zeros(init_vals.shape)
        mnl_estimator = MNLEstimator(self,
                                     mapping_res,
                                     PMLE,
                                     PMLE_lambda,
                                     zero_vector,
                                     split_param_vec,
                                     constrained_pos=constrained_pos)
        # Set the derivative functions for estimation
        mnl_estimator.set_derivatives()

        # Perform one final check on the length of the initial values
        mnl_estimator.check_length_of_initial_values(init_vals)

        # Get the estimation results
        estimation_res = estimate(init_vals,
                                  mnl_estimator,
                                  method,
                                  loss_tol,
                                  gradient_tol,
                                  maxiter,
                                  print_res,
                                  just_point=just_point)

        if not just_point:
            # Store the estimation results
            self.store_fit_results(estimation_res)

            return None
        else:
            return estimation_res



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 30 11:57:29 2016

@module:    nested_coice_calcs.py
@name:      Nested Choice Model Calculations
@author:    Timothy Brathwaite
@summary:   Contains generic functions necessary for calculating 'nested'
            choice probabilities and for estimating the 'nested' choice models.
"""
import numpy as np
from scipy.sparse import issparse

# Define the boundary values which are not to be exceeded ducing computation
min_exponent_val = -700
max_exponent_val = 700

max_comp_value = 1e300
min_comp_value = 1e-300


def calc_nested_probs(nest_coefs,
                      index_coefs,
                      design,
                      rows_to_obs,
                      rows_to_nests,
                      chosen_row_to_obs=None,
                      return_type="long_probs",
                      *args,
                      **kwargs):
    """
    Parameters
    ----------
    nest_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each nesting coefficient being estimated. If 2D, should
        have 1 column for each set of nesting coefficients being used to
        predict the probabilities of each alternative being chosen. There
        should be one row per nesting coefficient. Elements denote the inverse
        of the scale coefficients for each of the lower level nests.
    index_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated (i.e.
        num_features). If 2D, should have 1 column for each set of coefficients
        being used to predict the probabilities of each alternative being
        chosen. There should be one row per index coefficient.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_nests : 2D scipy sparse array.
        There should with one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).
    chosen_row_to_obs : 2D scipy sparse array, or None, optional.
        There should be one row per observation per available alternative and
        one column per observation. This matrix indicates, for each observation
        (on the columns), which rows of the design matrix were the realized
        outcome. If an array is passed then an array of shape
        (num_observations,) can be returned and each element will be the
        probability of the realized outcome of the given observation.
        Default == None.
    return_type : str, optional.
        Indicates what object(s) are to be returned from the function. Valid
        values are: `['long_probs', 'chosen_probs', 'long_and_chosen_probs',
        'all_prob_dict']`. If `long_probs`, the long format probabilities (a 1D
        numpy array with one element per observation per available alternative)
        will be returned. If `chosen_probs`, a 1D numpy array with one element
        per observation will be returned, where the values are the
        probabilities of the chosen alternative for the given observation. If
        `long_and_chosen_probs`, a tuple of chosen_probs and long_probs will be
        returned. If `all_prob_dict`, a dictionary will be returned. The values
        will all be 1D numpy arrays of probabilities dictated by the value's
        corresponding key. The keys will be `long_probs`, `nest_choice_probs`,
        `prob_given_nest`, and `chosen_probs`. If chosen_row_to_obs is None,
        then `chosen_probs` will be None. If `chosen_row_to_obs` is passed,
        then `chosen_probs` will be a 1D array as described above.
        `nest_choice_probs` is of the same shape as `rows_to_nests` and it
        denotes the probability of each individual choosing each of the
        possible nests. `prob_given_nest` is of the same shape as `long_probs`
        and it denotes the probability of the individual associated with a
        given row choosing the alternative associated with that row, given that
        the individual chooses the nest that contains the given alternative.
        Default == `long_probs`.

    Returns
    -------
    See above for documentation of the `return_type` kwarg.
    """
    # Check for 2D index coefficients or nesting coefficients
    try:
        assert len(index_coefs.shape) <= 2
        assert (len(index_coefs.shape) == 1) or (index_coefs.shape[1] == 1)
        assert len(nest_coefs.shape) <= 2
        assert (len(nest_coefs.shape) == 1) or (nest_coefs.shape[1] == 1)
    except AssertionError:
        msg = "Support for 2D index_coefs or nest_coefs not yet implemented."
        raise NotImplementedError(msg)

    # Check for kwarg validity
    valid_return_types = ['long_probs',
                          'chosen_probs',
                          'long_and_chosen_probs',
                          'all_prob_dict']
    if return_type not in valid_return_types:
        msg = "return_type must be one of the following values: "
        raise ValueError(msg + str(valid_return_types))

    chosen_probs_needed = ['chosen_probs', 'long_and_chosen_probs']
    if chosen_row_to_obs is None and return_type in chosen_probs_needed:
        msg = "chosen_row_to_obs is None AND return_type in {}."
        raise ValueError(msg.format(chosen_probs_needed) +
                         "\nThis is invalid.")

    # Calculate the index for each alternative for each individual, V = X*beta
    index_vals = design.dot(index_coefs)

    # Get the long format nest parameters for each row of the design matrix
    long_nest_coefs = rows_to_nests.dot(nest_coefs)

    # Calculate the scaled index values (index / nest_param = V / lambda)
    scaled_index = index_vals / long_nest_coefs

    # Guard against overflow
    pos_inf_idx = np.isposinf(scaled_index)
    neg_inf_idx = np.isneginf(scaled_index)
    scaled_index[pos_inf_idx] = max_comp_value
    scaled_index[neg_inf_idx] = -1 * max_comp_value

    # Calculate the e^(scaled-index) = exp(V / lambda)
    exp_scaled_index = np.exp(scaled_index)

    # Guard against overflow
    inf_idx = np.isposinf(exp_scaled_index)
    exp_scaled_index[inf_idx] = max_comp_value
    # Guard against underflow. Note that I'm not sure this is the best place or
    # best way to perform such guarding. If all of an observations indices
    # suffer underflow, then we'll have 0 / 0 when calculating the
    # probabilities and I should use L'Hopital's rule to get the correct
    # probability. However, replacing underflowed values here may result in
    # incorrectly assigning probabilities of either zero for all alternatives
    # or 1 / num_alternatives for all alternatives.
    zero_idx = (exp_scaled_index == 0)
    exp_scaled_index[zero_idx] = min_comp_value

    # Calculate the log-sum for each nest, for each observation. Note that the
    # "*" is used to compute the dot product between the mapping matrix which
    # is a scipy.sparse matrix and the second term which is a scipy sparse
    # matrix. Note the dimensions of ind_log_sums_per_nest are (obs, nests).
    # Calculates sum _{j \in C_m} exp(V_{ij} / \lambda_m) for each nest m.
    ind_exp_sums_per_nest = (rows_to_obs.T *
                             rows_to_nests.multiply(exp_scaled_index[:, None]))
    # Ensure that ind_exp_sums_per_nest is an ndarray
    if isinstance(ind_exp_sums_per_nest, np.matrixlib.defmatrix.matrix):
        ind_exp_sums_per_nest = np.asarray(ind_exp_sums_per_nest)
    elif issparse(ind_exp_sums_per_nest):
        ind_exp_sums_per_nest = ind_exp_sums_per_nest.toarray()
    # Guard against overflow
    inf_idx = np.isposinf(ind_exp_sums_per_nest)
    ind_exp_sums_per_nest[inf_idx] = max_comp_value

    # Get the long-format representation of ind_log_sums_per_nest. Each row
    # will have two columns, one for each nest. The entries of the matrix will
    # be the log-sum for each nest, for the individual associated with the
    # given row. The "*" is used to perform the dot product since rows_to_obs
    # is a sparse matrix & ind_exp_sums_per_nest is a dense numpy matrix.
    long_exp_sums_per_nest = rows_to_obs.dot(ind_exp_sums_per_nest)
    if isinstance(long_exp_sums_per_nest, np.matrixlib.defmatrix.matrix):
        long_exp_sums_per_nest = np.asarray(long_exp_sums_per_nest)

    # Get the relevant log-sum for each row of the long-format data
    # Note the .A converts the numpy matrix into a numpy array
    # This is sum _{j \in C_m} exp(V_{ij} / \lambda_m) for the nest
    # belonging to each row
    long_exp_sums = (rows_to_nests.multiply(long_exp_sums_per_nest)
                                  .sum(axis=1)
                                  .A).ravel()

    # Get the denominators for each individual
    ind_denom = (np.power(ind_exp_sums_per_nest,
                          nest_coefs[None, :])
                   .sum(axis=1))
    # Guard against overflow and underflow
    inf_idx = np.isposinf(ind_denom)
    ind_denom[inf_idx] = max_comp_value

    zero_idx = (ind_denom == 0)
    ind_denom[zero_idx] = min_comp_value

    # Get the long format denominators.
    long_denom = rows_to_obs.dot(ind_denom)
    # Ensure that long_denom is 1D.
    long_denom.ravel()

    # Get the long format numerators
    long_numerators = (exp_scaled_index *
                       np.power(long_exp_sums,
                                (long_nest_coefs - 1)))
    # Guard agains overflow and underflow
    inf_idx = np.isposinf(long_numerators)
    long_numerators[inf_idx] = max_comp_value

    zero_idx = (long_numerators == 0)
    long_numerators[zero_idx] = min_comp_value

    # Calculate and return the long-format probabilities
    long_probs = (long_numerators / long_denom).ravel()
    # Guard against underflow
    long_probs[np.where(long_probs == 0)] = min_comp_value

    # If desired, isolate the probabilities of the chosen alternatives
    if chosen_row_to_obs is None:
        chosen_probs = None
    else:
        # chosen_probs will be of shape (num_observations,)
        chosen_probs = (chosen_row_to_obs.transpose()
                                         .dot(long_probs))
        chosen_probs = np.asarray(chosen_probs).ravel()

    # Return the long form and chosen probabilities if desired
    if return_type == 'long_and_chosen_probs':
        return chosen_probs, long_probs
    # If working with predictions, return just the long form probabilities
    elif return_type == 'long_probs':
        return long_probs
    # If estimating the model and storing fitted probabilities or testing the
    # model on data for which we know the chosen alternative, just return the
    # chosen probabilities.
    elif return_type == 'chosen_probs':
        return chosen_probs
    # If we want all the factors of the probability (e.g. as when calculating
    # the gradient)
    elif return_type == 'all_prob_dict':
        # Create the dictionary of the various probabilities to be returned
        prob_dict = {}
        prob_dict["long_probs"] = long_probs
        prob_dict["chosen_probs"] = chosen_probs

        # Calculate the 'prob_given_nest' array
        prob_given_nest = exp_scaled_index / long_exp_sums
        # Guard against underflow
        zero_idx = (prob_given_nest == 0)
        prob_given_nest[zero_idx] = min_comp_value

        # Calculate the 'nest_choice_probs'. Note ind_denom is a matrix with
        # shape (num_obs, 1) so no need to explicitly broadcast
        nest_choice_probs = (np.power(ind_exp_sums_per_nest,
                                      nest_coefs[None, :]) /
                             ind_denom[:, None])
        # Guard against underflow
        zero_idx = (nest_choice_probs == 0)
        nest_choice_probs[zero_idx] = min_comp_value
        # Return dictionary.
        # Note the ".A" converts the numpy matrix into a numpy array
        prob_dict["prob_given_nest"] = prob_given_nest
        prob_dict["nest_choice_probs"] = nest_choice_probs
        prob_dict["ind_sums_per_nest"] = ind_exp_sums_per_nest

        return prob_dict


def calc_nested_log_likelihood(nest_coefs,
                               index_coefs,
                               design,
                               rows_to_obs,
                               rows_to_nests,
                               choice_vector,
                               ridge=None,
                               weights=None,
                               *args,
                               **kwargs):
    """
    Parameters
    ----------
    nest_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each nesting coefficient being estimated. If 2D, should
        have 1 column for each set of nesting coefficients being used to
        predict the probabilities of each alternative being chosen. There
        should be one row per nesting coefficient. Elements denote the inverse
        of the scale coefficients for each of the lower level nests.
    index_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated
        (i.e. num_features). If 2D, should have 1 column for each set of
        coefficients being used to predict the probabilities of choosing each
        alternative. There should be one row per index coefficient.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_nests : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).
    choice_vector : 1D ndarray.
        All elements should be either ones or zeros. There should be one row
        per observation per available alternative for the given observation.
        Elements denote the alternative which is chosen by the given
        observation with a 1 and a zero otherwise.
    ridge : int, float, long, or None, optional.
        Determines whether or not ridge regression is performed. If an int,
        float or long is passed, then that scalar determines the ridge penalty
        for the optimization. Default = None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    Returns
    -------
    log_likelihood : float.
        The log likelihood of the nested logit model. Includes ridge penalty if
        a penalized regression is being performed.
    """
    # Calculate the probability of each individual choosing each available
    # alternative for that individual.
    long_probs = calc_nested_probs(nest_coefs,
                                   index_coefs,
                                   design,
                                   rows_to_obs,
                                   rows_to_nests,
                                   return_type='long_probs')

    # Calculate the weights for the sample
    if weights is None:
        weights = 1

    # Calculate the log likelihood
    log_likelihood = choice_vector.dot(weights * np.log(long_probs))

    if ridge is None:
        return log_likelihood
    else:
        # Note that the 1.0 is used since the 'null' nest coefficient is equal
        # to 1.0.
        params = np.concatenate(((nest_coefs - 1.0), index_coefs), axis=0)

        return log_likelihood - ridge * np.square(params).sum()


# Create a function to create the various arrays that are needed when
# calculating the gradient of the nested logit model
def prep_vectors_for_gradient(nest_coefs,
                              index_coefs,
                              design,
                              choice_vec,
                              rows_to_obs,
                              rows_to_nests,
                              *args,
                              **kwargs):
    """
    Parameters
    ----------
    nest_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each nesting coefficient being estimated. If 2D, should
        have 1 column for each set of nesting coefficients being used to
        predict the probabilities of each alternative being chosen. There
        should be one row per nesting coefficient. Elements denote the inverse
        of the scale coefficients for each of the lower level nests. Note, this
        is NOT THE LOGIT of the inverse of the scale coefficients.
    index_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated
        (i.e. num_features). If 2D, should have 1 column for each set of
        coefficients being used to predict the probabilities of choosing each
        alternative. There should be one row per index coefficient.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated.
        All elements should be ints, floats, or longs.
    choice_vec : 1D ndarray.
        All elements should by ints, floats, or longs. Each element represents
        whether the individual associated with the given row chose the
        alternative associated with the given row. Should have the same number
        of rows as `design`.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_nests : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).

    Returns
    -------
    desired_arrays : dict.
        Will contain the arrays necessary for calculating the gradient of the
        nested logit log-likelihood. The keys will be:
            `["long_nest_params", "scaled_y", "long_chosen_nest",
              "obs_to_chosen_nests", "p_tilde_given_nest", "long_probs",
              "prob_given_nest", "nest_choice_probs", "ind_sums_per_nest"]`
    """
    # Create the "long_nest_parameters" which is an array with one element per
    # alternative per observation, where each element is the nest parameter for
    # the alternative corresponding to the given row
    long_nest_params = (rows_to_nests.multiply(nest_coefs[None, :])
                                     .sum(axis=1)
                                     .A
                                     .ravel())

    # Calculate y-tilde
    scaled_y = choice_vec / long_nest_params
    # Guard against overflow
    inf_index = np.isinf(scaled_y)
    scaled_y[inf_index] = max_comp_value

    # Determine which nest was chosen by each row's individual.
    # Resulting matrix has shape (num_rows, num_nests)
    obs_to_chosen_nests = (rows_to_obs.T *
                           rows_to_nests.multiply(choice_vec[:, None])).A
    row_to_chosen_nest = rows_to_obs * obs_to_chosen_nests
    # Determine whether the given row is part of the nest that was chosen
    long_chosen_nest = (rows_to_nests.multiply(row_to_chosen_nest)
                                     .sum(axis=1)
                                     .A
                                     .ravel())

    # Get the various probabilities
    prob_dict = calc_nested_probs(nest_coefs,
                                  index_coefs,
                                  design,
                                  rows_to_obs,
                                  rows_to_nests,
                                  return_type='all_prob_dict')

    # Calculate p_tilde_ij_given_nest
    p_tilde_row_given_nest = (prob_dict["prob_given_nest"] *
                              long_chosen_nest /
                              long_nest_params)
    # Guard against overflow
    inf_index = np.isinf(p_tilde_row_given_nest)
    p_tilde_row_given_nest[inf_index] = max_comp_value

    # Return all the desired matrices and arrays
    desired_arrays = {}
    desired_arrays["long_nest_params"] = long_nest_params.ravel()
    desired_arrays["scaled_y"] = scaled_y.ravel()
    desired_arrays["long_chosen_nest"] = long_chosen_nest
    desired_arrays["obs_to_chosen_nests"] = obs_to_chosen_nests
    desired_arrays["p_tilde_given_nest"] = p_tilde_row_given_nest
    desired_arrays["long_probs"] = prob_dict["long_probs"]
    desired_arrays["prob_given_nest"] = prob_dict["prob_given_nest"]
    desired_arrays["nest_choice_probs"] = prob_dict["nest_choice_probs"]
    desired_arrays["ind_sums_per_nest"] = prob_dict["ind_sums_per_nest"]

    return desired_arrays


# Define a function to convert the transformed nest coefs
# (i.e. the logit of the nest coefficients) to the
# 'standard' nest coefficients
def naturalize_nest_coefs(nest_coef_estimates):
    """
    Parameters
    ----------
    nest_coef_estimates : 1D ndarray.
        Should contain the estimated logit's
        (`ln[nest_coefs / (1 - nest_coefs)]`) of the true nest coefficients.
        All values should be ints, floats, or longs.

    Returns
    -------
    nest_coefs : 1D ndarray.
        Will contain the 'natural' nest coefficients:
        `1.0 / (1.0 + exp(-nest_coef_estimates))`.
    """
    # Calculate the exponential term of the
    # logistic transformation
    exp_term = np.exp(-1 * nest_coef_estimates)

    # Guard against_overflow
    inf_idx = np.isinf(exp_term)
    exp_term[inf_idx] = max_comp_value

    # Calculate the 'natural' nest coefficients
    nest_coefs = 1.0 / (1.0 + exp_term)

    # Guard against underflow
    zero_idx = (nest_coefs == 0)
    nest_coefs[zero_idx] = min_comp_value

    return nest_coefs


# Create the actual function used to calculate the gradient
def calc_nested_gradient(orig_nest_coefs,
                         index_coefs,
                         design,
                         choice_vec,
                         rows_to_obs,
                         rows_to_nests,
                         ridge=None,
                         weights=None,
                         use_jacobian=True,
                         *args,
                         **kwargs):
    """
    Parameters
    ----------
    orig_nest_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each nesting coefficient being estimated. If 2D, should
        have 1 column for each set of nesting coefficients being used to
        predict the probabilities of each alternative being chosen. There
        should be one row per nesting coefficient. Elements denote the logit of
        the inverse of the scale coefficients for each lower level nests.
    index_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated
        (i.e. num_features). If 2D, should have 1 column for each set of
        coefficients being used to predict the probabilities of choosing each
        alternative. There should be one row per index coefficient.
    design : 2D ndarray.
       There should be one row per observation per available alternative. There
       should be one column per utility coefficient being estimated. All
       elements should be ints, floats, or longs.
    choice_vec : 1D ndarray.
        All elements should by ints, floats, or longs. Each element represents
        whether the individual associated with the given row chose the
        alternative associated with the given row.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_nests : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).
    ridge : int, float, long, or None, optional.
        Determines whether or not ridge regression is performed. If an int,
        float or long is passed, then that scalar determines the ridge penalty
        for the optimization. Default `== None`.
    weights : 1D ndarray or None.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    use_jacobian : bool, optional.
        Determines whether or not the jacobian will be used when calculating
        the gradient. When performing model estimation, `use_jacobian` should
        be `True` if the values being estimated are actually the logit of the
        nest coefficients. Default `== True`.

    Returns
    -------
    gradient : 1D numpy array.
       The gradient of the log-likelihood with respect to the given nest
       coefficients and index coefficients.
    """
    # Calculate the weights for the sample
    if weights is None:
        weights = np.ones(design.shape[0])
    weights_per_obs = np.max(rows_to_obs.toarray() * weights[:, None], axis=0)

    # Transform the nest coefficients into their "always positive" versions
    nest_coefs = naturalize_nest_coefs(orig_nest_coefs)

    # Get the vectors and matrices needed to calculate the gradient
    vector_dict = prep_vectors_for_gradient(nest_coefs,
                                            index_coefs,
                                            design,
                                            choice_vec,
                                            rows_to_obs,
                                            rows_to_nests)

    # Calculate the index for each alternative for each person
    sys_utility = design.dot(index_coefs)

    # Calculate w_ij
    long_w = sys_utility / vector_dict["long_nest_params"]
    # Guard against overflow
    inf_index = np.isposinf(long_w)
    long_w[inf_index] = max_comp_value

    ##########
    # Calculate d_log_likelihood_d_nest_params
    ##########
    # Calculate the term that onlny depends on nest level values
    log_exp_sums = np.log(vector_dict["ind_sums_per_nest"])
    # Guard against overflow
    log_exp_sums[np.isneginf(log_exp_sums)] = -1 * max_comp_value

    # Calculate the first term of the derivative of the log-liikelihood
    # with respect to the nest parameters
    nest_gradient_term_1 = ((vector_dict["obs_to_chosen_nests"] -
                             vector_dict["nest_choice_probs"]) *
                            log_exp_sums *
                            weights_per_obs[:, None]).sum(axis=0)

    # Calculate the second term of the derivative of the log-liikelihood
    # with respect to the nest parameters
    half_deriv = ((vector_dict["long_probs"] -
                   vector_dict["long_chosen_nest"] *
                   vector_dict["prob_given_nest"]) *
                  long_w *
                  weights)
    nest_gradient_term_2 = (rows_to_nests.transpose()
                                         .dot(half_deriv)[:, None]).ravel()

    # Calculate the third term of the derivative of the log-likelihood
    # with respect to the nest parameters
    nest_gradient_term_3a = (choice_vec -
                             vector_dict["long_chosen_nest"] *
                             vector_dict["prob_given_nest"])
    nest_gradient_term_3b = ((-1 * nest_gradient_term_3a * long_w * weights) /
                             vector_dict["long_nest_params"])
    # Guard against overflow
    inf_idx = np.isposinf(nest_gradient_term_3b)
    nest_gradient_term_3b[inf_idx] = max_comp_value

    neg_inf_idx = np.isneginf(nest_gradient_term_3b)
    nest_gradient_term_3b[neg_inf_idx] = -1 * max_comp_value

    # Get the nest-wide version of this piece of the gradient
    nest_gradient_term_3 = (rows_to_nests.transpose()
                                         .dot(nest_gradient_term_3b)).ravel()

    # Combine the two terms. Note the "nest_coefs * (1 - nest_coefs)" is due to
    # the fact that we're estimating the logit of the nest coefficients instead
    # of the nest coefficient itself. We therefore need to multiply by
    # d_nest_coef_d_estimated_variable to get the correct gradient.
    # d_nest_coef_d_estimated_variable == nest_coefs * (1 - nest_coefs).
    if use_jacobian:
        jacobian = nest_coefs * (1.0 - nest_coefs)
    else:
        jacobian = 1
    nest_gradient = ((nest_gradient_term_1 +
                      nest_gradient_term_2 +
                      nest_gradient_term_3) *
                     jacobian)[None, :]

    ##########
    # Calculate d_loglikelihood_d_beta
    ##########
    beta_gradient_term_1 = ((vector_dict["scaled_y"] -
                             vector_dict["p_tilde_given_nest"] +
                             vector_dict["p_tilde_given_nest"] *
                             vector_dict["long_nest_params"] -
                             vector_dict["long_probs"]) *
                            weights)[None, :]
    #####
    # Calculate the derivative with respect to beta
    #####
    beta_gradient = beta_gradient_term_1.dot(design)

    #####
    # Combine the gradient pieces and account for ridge parameter
    #####
    gradient = np.concatenate((nest_gradient, beta_gradient), axis=1).ravel()

    if ridge is not None:
        # Note that the 20 is used in place of 'infinity' since I would really
        # like to specify the expected value of the nest coefficient to 1, but
        # that would make the logit of the nest parameter infinity. Instead I
        # use 20 as a close enough value-- (1 + exp(-20))**-1 is approx. 1.
        params = np.concatenate(((20 - orig_nest_coefs), index_coefs), axis=0)

        gradient -= 2 * ridge * params

    return gradient


# Define a function that will calculate the BHHH approximation of the hessian
# This is essentially the sum over all individuals, of the outer product of
# the gradient with itself.
def calc_bhhh_hessian_approximation(orig_nest_coefs,
                                    index_coefs,
                                    design,
                                    choice_vec,
                                    rows_to_obs,
                                    rows_to_nests,
                                    ridge=None,
                                    weights=None,
                                    use_jacobian=True,
                                    *args,
                                    **kwargs):
    """
    Parameters
    ----------
    orig_nest_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each nesting coefficient being estimated. If 2D, should
        have 1 column for each set of nesting coefficients being used to
        predict the probabilities of each alternative being chosen. There
        should be one row per nesting coefficient. Elements denote the inverse
        of the scale coefficients for each of the lower level nests.
    index_coefs : 1D or 2D ndarray.
        All elements should by ints, floats, or longs. If 1D, should have 1
        element for each utility coefficient being estimated
        (i.e. num_features). If 2D, should have 1 column for each set of
        coefficients being used to predict the probabilities of choosing each
        alternative. There should be one row per index coefficient.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    choice_vec : 1D ndarray.
        All elements should by ints, floats, or longs. Each element represents
        whether the individual associated with the given row chose the
        alternative associated with the given row.
    rows_to_obs : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per observation. This matrix maps the rows of the design
        matrix to the unique observations (on the columns).
    rows_to_nests : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).
    ridge : int, float, long, or None, optional.
        Determines whether or not ridge regression is performed. If an int,
        float or long is passed, then that scalar determines the ridge penalty
        for the optimization. Default == None. Note that if this parameter is
        passed, the values of the BHHH matrix MAY BE INCORRECT since it is not
        100% clear how penalization affects the information matrix.
    use_jacobian : bool, optional.
        Determines whether or not the jacobian will be used when calculating
        the gradient. When performing model estimation, `use_jacobian` should
        be `True` if the values that are actually being estimated are the
        logit of the nest coefficients. Default `== False`.

    Returns
    -------
    bhhh_matrix : 2D ndarray.
       The negative of the sum of the outer products of the gradient of the
       log-likelihood function for each observation.
    """
    # Calculate the weights for the sample
    if weights is None:
        weights = np.ones(design.shape[0])
    weights_per_obs = np.max(rows_to_obs.toarray() * weights[:, None], axis=0)

    # Transform the nest coefficients into their "always positive" versions
    nest_coefs = naturalize_nest_coefs(orig_nest_coefs)

    # Get the vectors and matrices needed to calculate the gradient
    vector_dict = prep_vectors_for_gradient(nest_coefs,
                                            index_coefs,
                                            design,
                                            choice_vec,
                                            rows_to_obs,
                                            rows_to_nests)

    # Calculate the index for each alternative for each person
    sys_utility = design.dot(index_coefs)

    # Calculate w_ij
    long_w = sys_utility / vector_dict["long_nest_params"]
    # Guard against overflow
    inf_index = np.isposinf(long_w)
    long_w[inf_index] = max_comp_value

    ##########
    # Calculate d_log_likelihood_d_nest_params
    ##########
    # Calculate the term that only depends on nest level values
    log_exp_sums = np.log(vector_dict["ind_sums_per_nest"])
    # Guard against overflow
    log_exp_sums[np.isneginf(log_exp_sums)] = -1 * max_comp_value

    # Calculate the first term of the derivative of the log-liikelihood
    # with respect to the nest parameters. Note we do not sum this object since
    # we want the values at the 'individual' level, which they already are.
    nest_gradient_term_1 = ((vector_dict["obs_to_chosen_nests"] -
                             vector_dict["nest_choice_probs"]) *
                            log_exp_sums)

    # Calculate the second term of the derivative of the log-liikelihood
    # with respect to the nest parameters
    half_deriv = ((vector_dict["long_probs"] -
                   vector_dict["long_chosen_nest"] *
                   vector_dict["prob_given_nest"]) *
                  long_w)[:, None]
    # "Spread out" the second term across the appropriate nests
    spread_half_deriv = rows_to_nests.multiply(half_deriv)
    # Aggregate the spread out half-derivatives to the individual level
    # This object should have shape (num_obs, num_nests)
    nest_gradient_term_2 = rows_to_obs.transpose().dot(spread_half_deriv).A

    # Calculate the third term of the derivative of the log-likelihood
    # with respect to the nest parameters
    nest_gradient_term_3a = (choice_vec -
                             vector_dict["long_chosen_nest"] *
                             vector_dict["prob_given_nest"])

    nest_gradient_term_3b = ((-1 * nest_gradient_term_3a * long_w) /
                             vector_dict["long_nest_params"])

    # Guard against overflow
    inf_idx = np.isposinf(nest_gradient_term_3b)
    nest_gradient_term_3b[inf_idx] = max_comp_value

    neg_inf_idx = np.isneginf(nest_gradient_term_3b)
    nest_gradient_term_3b[neg_inf_idx] = -1 * max_comp_value

    # Get the nest-wide version of this piece of the gradient
    spread_out_term_3b = rows_to_nests.multiply(nest_gradient_term_3b[:, None])
    nest_gradient_term_3 = rows_to_obs.transpose().dot(spread_out_term_3b).A

    # Combine the terms. Note the "nest_coefs * (1 - nest_coefs)" is due to the
    # fact that we're estimating the logit of the nest coefficients instead of
    # the nest coefficient itself. We therefore need to multiply by
    # d_nest_coef_d_estimated_variable to get the correct gradient.
    # d_nest_coef_d_estimated_variable == nest_coefs * (1 - nest_coefs).
    # As with the various nest_gradient_terms, the nest_gradient should be of
    # shape (num_obs, num_nests)
    if use_jacobian:
        jacobian = (nest_coefs * (1.0 - nest_coefs))[None, :]
    else:
        jacobian = 1

    nest_gradient = ((nest_gradient_term_1 +
                      nest_gradient_term_2 +
                      nest_gradient_term_3) *
                     jacobian)

    ##########
    # Calculate d_loglikelihood_d_beta
    ##########
    beta_gradient_term_1 = (vector_dict["scaled_y"] -
                            vector_dict["p_tilde_given_nest"] +
                            vector_dict["p_tilde_given_nest"] *
                            vector_dict["long_nest_params"] -
                            vector_dict["long_probs"])[:, None]
    #####
    # Calculate the derivative with respect to beta
    #####
    beta_gradient = rows_to_obs.T.dot(beta_gradient_term_1 * design)

    #####
    # Combine the gradient pieces
    #####
    gradient_matrix = np.concatenate((nest_gradient, beta_gradient), axis=1)

    #####
    # Compute and return the outer product of each row of the gradient
    # with itself. Then sum these individual matrices together. The line below
    # does the same computation just with less memory and time.
    bhhh_matrix =\
        gradient_matrix.T.dot(weights_per_obs[:, None] * gradient_matrix)

    if ridge is not None:
        # The rational behind adding 2 * ridge is that the information
        # matrix should approximate the hessian and in the hessian we subtract
        # 2 * ridge at the end. We add 2 * ridge here, since we will multiply
        # by negative one afterwards. I don't know if this is the correct way
        # to calculate the Fisher Information Matrix in ridge regression
        # models.
        bhhh_matrix += 2 * ridge

    # Note the "-1" is because we are approximating the Fisher information
    # matrix which has a negative one in the front of it?
    # Note that if we were using the bhhh_matrix to calculate the robust
    # covariance matrix, then we would not multiply by negative one here.
    return -1 * bhhh_matrix
